import pandas as pd
import numpy as np
import operator
from packaging import version
from .engineobj import SqPandasEngine
from suzieq.utils import humanize_timestamp


class DeviceObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'device'

    def get(self, **kwargs):
        """Get the information requested"""
        view = kwargs.get('view', self.iobj.view)
        columns = kwargs.get('columns', ['default'])
        addnl_fields = kwargs.pop('addnl_fields', [])
        user_query = kwargs.pop('query_str', '')
        status = kwargs.pop('status', '')
        os_version = kwargs.pop('version', '')
        vendor = kwargs.get('vendor', '')
        model = kwargs.get('model', '')
        os = kwargs.get('os', '')

        drop_cols = []

        if 'active' not in addnl_fields+columns and columns != ['*']:
            addnl_fields.append('active')
            drop_cols.append('active')

        # os is not included in the default column list. Why? I was dumb
        if (columns == ['default'] and os) or (os and 'os' not in columns):
            addnl_fields.append('os')
            drop_cols.append('os')

        for col in ['namespace', 'hostname', 'status', 'address']:
            if (not ((columns == ['default']) or (columns == ['*'])) and
                    col not in columns):
                addnl_fields.append(col)
                drop_cols.append(col)

        if columns == ['*'] or 'uptime' in columns:
            if columns != ['*'] and 'bootupTimestamp' not in columns:
                addnl_fields.append('bootupTimestamp')
                drop_cols.append('bootupTimestamp')

        df = super().get(active_only=False, addnl_fields=addnl_fields,
                         **kwargs)
        if view == 'latest' and 'status' in df.columns:
            df['status'] = np.where(df.active, df['status'], 'dead')

        poller_df = self._get_table_sqobj('sqPoller').get(
            namespace=kwargs.get('namespace', []),
            hostname=kwargs.get('hostname', []),
            service='device',
            columns='namespace hostname status'.split())

        if not poller_df.empty:
            # Identify the address to namespace/hostname mapping
            addr_dict = {f"{x['namespace']}-{x['address']}": x['hostname']
                         for x in df[['namespace', 'address', 'hostname']]
                         .to_dict(orient='records')}

            poller_df['hostname'] = poller_df.apply(
                lambda x, y: y.get(f"{x['namespace']}-{x['hostname']}",
                                   x['hostname']),
                args=(addr_dict,), axis=1)

            poller_df = poller_df\
                .drop_duplicates(subset=['namespace', 'hostname'],
                                 keep='last') \
                .reset_index(drop=True)

            df = df.merge(poller_df, on=['namespace', 'hostname'],
                          how='outer', suffixes=['', '_y'])  \
                .fillna({'bootupTimestamp': 0, 'timestamp': 0,
                         'active': True}) \
                .fillna('N/A')

            df.status = np.where(
                (df['status_y'] != 0) & (df['status_y'] != 200) &
                (df['status'] == "N/A"),
                'neverpoll', df['status'])
            df = df[df.status != 'N/A']
            df.timestamp = np.where(df['timestamp'] == 0,
                                    df['timestamp_y'], df['timestamp'])
            if 'address' in df.columns:
                df.address = np.where(df['address'] == 'N/A', df['hostname'],
                                      df['address'])

            drop_cols.extend(['status_y', 'timestamp_y'])

            if 'uptime' in columns or columns == ['*']:
                uptime_cols = (df['timestamp'] -
                               humanize_timestamp(df['bootupTimestamp']*1000,
                               self.cfg.get('analyzer', {}).get('timezone', None)))
                uptime_cols = pd.to_timedelta(uptime_cols, unit='s')
                df.insert(len(df.columns)-1, 'uptime', uptime_cols)

        if df.empty:
            return df

        # The poller merge kills the filtering we did earlier, so redo:
        if status:
            df = df.loc[df.status.isin(status)]
        if vendor:
            df = df.loc[df.vendor.isin(vendor)]
        if model:
            df = df.loc[df.model.isin(model)]
        if os:
            df = df.loc[df.os.isin(os)]
        if os_version:
            opdict = {'>': operator.gt, '<': operator.lt, '>=': operator.ge,
                      '<=': operator.le, '=': operator.eq, '!=': operator.ne}
            op = operator.eq
            for elem in opdict:
                if os_version.startswith(elem):
                    os_version = os_version.replace(elem, '')
                    op = opdict[elem]
                    break

            df = df.loc[df.version.apply(
                lambda x: op(version.LegacyVersion(x), version.LegacyVersion(os_version)))]

        df = self._handle_user_query_str(df, user_query)

        # if poller has failed completely, Can mess up the order of columns
        cols = self.iobj.schema.get_display_fields(columns)
        if columns == ['default'] and 'timestamp' not in cols:
            cols.append('timestamp')
        if 'sqvers' in cols:
            cols.remove('sqvers')
        return df.drop(columns=drop_cols, errors='ignore')[cols]

    def summarize(self, **kwargs):
        """Summarize device information across namespace"""

        self._init_summarize(self.iobj.table, **kwargs)
        if self.summary_df.empty:
            return self.summary_df

        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
        ]

        self._summarize_on_add_with_query = [
            ('downDeviceCnt', 'status == "dead"', 'hostname', 'nunique'),
            ('unpolledDeviceCnt', 'status == "neverpoll"', 'hostname',
             'nunique'),
        ]

        self._summarize_on_add_list_or_count = [
            ('vendorCnt', 'vendor'),
            ('modelCnt', 'model'),
            ('archCnt', 'architecture'),
            ('versionCnt', 'version'),
        ]

        self.summary_df = self.iobj.humanize_fields(self.summary_df)

        self._summarize_on_add_stat = [
            ('upTimeStat', 'status != "neverpoll"', 'uptime')
        ]

        self._gen_summarize_data()
        self._post_summarize(check_empty_col='deviceCnt')
        return self.ns_df.convert_dtypes()



from suzieq.gui.guiutils import sq_gui_style, gui_get_df, SuzieqMainPages, display_title
from suzieq.sqobjects import get_sqobject
import numpy as np
import pandas as pd
import altair as alt
import streamlit as st
from dataclasses import dataclass, field, asdict
from typing import List


@dataclass
class XploreSessionState:
    page: str = SuzieqMainPages.XPLORE
    namespace: str = ''
    hostname: str = ''
    start_time: str = ''
    end_time: str = ''
    table: str = ''
    view: str = 'latest'
    query: str = ''
    columns:  List[str] = field(default_factory=list)
    uniq_clicked: str = '-'
    assert_clicked: bool = False
    state_changed: bool = False


def get_title():
    return 'Xplore'


@st.cache(ttl=90)
def xplore_run_summarize(sqobject, **kwargs):
    view = kwargs.pop('view', 'latest')
    stime = kwargs.pop('start_time', '')
    etime = kwargs.pop('end_time', '')
    df = sqobject(view=view, start_time=stime,
                  end_time=etime).summarize(**kwargs)
    # Leaving this commented to avoid future heartburn in case someone
    # tries to do this. It didn't fix the Timedelta being added to display
    # if not df.empty:
    #     if 'upTimeStat' in df.T.columns:
    #         df.T['upTimeStat'] = df.T.upTimeStat.apply(lambda x: [str(y)
    #                                                               for y in x])

    return df


def xplore_run_unique(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
    '''Compute the distribution counts for the dataframe provided'''
    column = kwargs.pop('columns', '')
    if not df.empty:
        if column == 'prefixlen' and 'prefixlen' not in df.columns:
            # Special handling for prefixlen
            df['prefixlen'] = df['prefix'].str.split('/').str[1]

        if df.apply(lambda x: isinstance(x[column], np.ndarray), axis=1).all():
            idf = df.explode(column).dropna(how='any')
        else:
            idf = df

        r = idf[column].value_counts()
        return (pd.DataFrame({column: r})
                .reset_index()
                .rename(columns={column: 'numRows',
                                 'index': column})
                .sort_values(by=['numRows'], ascending=False))


@st.cache(ttl=90)
def xplore_run_assert(sqobject, **kwargs):
    kwargs.pop('view', 'latest')
    stime = kwargs.pop('start_time', '')
    etime = kwargs.pop('end_time', '')
    df = sqobject(start_time=stime, end_time=etime) \
        .aver(status="fail", **kwargs)
    if not df.empty:
        df.rename(columns={'assert': 'status'},
                  inplace=True, errors='ignore')
    return df


def xplore_sidebar(state, sqobjs: dict):
    '''Draw appropriate sidebar for the page'''

    stime = state.start_time
    etime = state.end_time

    table_vals = [''] + sorted(list(sqobjs.keys()))

    if state.table:
        if isinstance(state.table, list):
            tblidx = table_vals.index(state.table[0])
        else:
            tblidx = table_vals.index(state.table)
    else:
        tblidx = 0  # Default starting table
    assert_val = state.assert_clicked
    view_idx = 1 if state.view == 'all' else 0

    devdf = gui_get_df(sqobjs['device'], columns=['namespace', 'hostname'])
    if devdf.empty:
        st.error('Unable to retrieve any namespace info')
        st.stop()

    namespaces = [""]
    namespaces.extend(sorted(devdf.namespace.unique().tolist()))
    if state.namespace:
        nsidx = namespaces.index(state.namespace)
    else:
        nsidx = 0
    with st.sidebar:
        with st.form('Xplore'):
            namespace = st.selectbox('Namespace',
                                     namespaces, key='xplore_namespace', index=nsidx)
            state.start_time = st.text_input('Start time',
                                             value=stime,
                                             key='xplore_stime')
            state.end_time = st.text_input('End time',
                                           value=etime,
                                           key='xplore_etime')

            table = st.selectbox(
                'Select Table to View', tuple(table_vals), key='xplore_table', index=tblidx)
            if table != state.table:
                # We need to reset the specific variables
                state.query = ''
                state.assert_clicked = False
                state.uniq_clicked = '-'
                state.columns = ['default']
                state.table = table

            view_vals = ('latest', 'all')
            if state.start_time and state.end_time:
                # We show everything thats happened when both times are specified
                view_idx = 1
            state.view = st.radio("View of Data", view_vals,
                                  index=view_idx, key='xplore_view')
            st.form_submit_button('Get', on_click=xplore_sync_state)

    if namespace != state.namespace:
        state.hostname = None
        state.namespace = namespace

    hostnames = [""]
    if state.namespace:
        hostlist = devdf.query(f'namespace=="{state.namespace}"') \
                        .hostname.unique().tolist()
    else:
        hostlist = devdf.hostname.unique().tolist()
    hostnames.extend(sorted(hostlist))
    if state.hostname:
        hostidx = hostnames.index(state.hostname)
    else:
        hostidx = 0
    state.hostname = st.sidebar.selectbox('Hostname',
                                          hostnames, index=hostidx,
                                          key='xplore_hostname',
                                          on_change=xplore_sync_state)

    if state.table:
        tables_obj = get_sqobject('tables')(start_time=state.state_time,
                                            end_time=state.end_time,
                                            view=state.view)
        fields = tables_obj.describe(table=state.table)
        colist = sorted((filter(lambda x: x not in ['index', 'sqvers'],
                                fields.name.tolist())))
        columns = st.sidebar.multiselect('Pick columns',
                                         ['default', 'all'] + colist,
                                         key='xplore_columns',
                                         default=state.columns)
        if ('default' in columns or 'all' in columns) and len(columns) == 1:
            col_sel_val = True
        else:
            col_sel_val = False

        col_ok = st.sidebar.checkbox('Column Selection Done',
                                     key='xplore_col_done',
                                     value=col_sel_val)
        if not col_ok:
            columns = ['default']
    else:
        col_ok = True
        columns = ['default']

    if not columns:
        columns = ['default']

    state.columns = columns
    if state.table in ['interfaces', 'ospf', 'bgp', 'evpnVni']:
        state.assert_clicked = st.sidebar.checkbox(
            'Run Assert', value=assert_val,  key='xplore_assert', on_change=xplore_sync_state)
    else:
        state.assert_clicked = False

    if not col_ok:
        st.experimental_set_query_params(**asdict(state))
        st.stop()
    if ('default' in columns or 'all' in columns) and len(columns) != 1:
        st.error('Cannot select default/all with any other columns')
        st.experimental_set_query_params(**asdict(state))
        st.stop()
    elif not columns:
        st.error('Columns cannot be empty')
        st.experimental_set_query_params(**asdict(state))
        st.stop()

    state.query = st.sidebar.text_input(
        'Filter results with pandas query', value=state.query,
        key='xplore_query', on_change=xplore_sync_state)

    st.sidebar.markdown(
        "[query syntax help](https://suzieq.readthedocs.io/en/latest/pandas-query-examples/)")

    if columns == ['all']:
        columns = ['*']

    if state.table:
        col_expander = st.sidebar.expander('Column Names', expanded=False)
        with col_expander:
            st.subheader(f'{state.table} column names')
            st.table(tables_obj.describe(table=state.table)
                     .query('name != "sqvers"')
                     .reset_index(drop=True).astype(str).style)


def page_work(state_container):
    '''The main workhorse routine for the Xplore page'''

    if 'xploreSessionState' not in state_container:
        state_container['xploreSessionState'] = None

    state = state_container.xploreSessionState

    if 'xplore_df' not in state_container:
        state_container['xplore_df'] = None

    url_params = st.experimental_get_query_params()
    page = url_params.pop('page', '')
    if not state and get_title() in page:
        if url_params and not all(not x for x in url_params.values()):
            for key in url_params:
                if key == 'columns':
                    # This needs to be a list
                    continue
                val = url_params.get(key, '')
                if isinstance(val, list):
                    val = val[0]
                    url_params[key] = val
                if key in ['assert_clicked', 'run_on_start']:
                    if val == 'True':
                        url_params[key] = True
                    else:
                        url_params[key] = False
            state = XploreSessionState(**url_params)
            state_container.xploreSessionState = state
        else:
            state = XploreSessionState()
            state_container.xploreSessionState = state
    elif not state:
        state = XploreSessionState()
        state_container.xploreSessionState = state

    sqobjs = state_container.sqobjs

    # All the user input is preserved in the state vars
    xplore_sidebar(state, sqobjs)
    xplore_run()


def xplore_sync_state():
    '''Update session state with latest widget state'''
    wsstate = st.session_state
    state = wsstate.xploreSessionState

    state_changed = False
    if wsstate.xplore_namespace != state.namespace:
        state.namespace = wsstate.xplore_namespace
        state_changed = True

    if wsstate.xplore_hostname != state.hostname:
        state.hostname = wsstate.xplore_hostname
        state_changed = True
    if wsstate.xplore_stime != state.start_time:
        state.start_time = wsstate.xplore_stime
        state_changed = True
    if wsstate.xplore_etime != state.end_time:
        state.end_time = wsstate.xplore_etime
        state_changed = True
    if wsstate.xplore_view != state.view:
        state.view = wsstate.xplore_view
        state_changed = True
    xplore_col_done = wsstate.get('xplore_col_done', False)
    if xplore_col_done and (wsstate.xplore_columns != state.columns):
        state.columns = wsstate.xplore_columns
        state_changed = True
    if wsstate.xplore_query != state.query:
        state.query = wsstate.xplore_query
        state_changed = True
    assert_clicked = wsstate.get('xplore_assert', '')
    if assert_clicked and (assert_clicked != state.assert_clicked):
        state.assert_clicked = wsstate.xplore_assert
        state_changed = True
    uniq_sel = wsstate.get('xplore_uniq_col', '')
    if uniq_sel and (state.uniq_clicked != uniq_sel):
        state.uniq_clicked = uniq_sel
        state_changed = True

    if wsstate.xplore_table != state.table:
        # Reset all dependent vars when table changes
        state.table = wsstate.xplore_table
        state.query = ''
        state.assert_clicked = False
        state.uniq_clicked = '-'
        state.columns = ['default']
        state_changed = True

    state.state_changed = state_changed


def xplore_create_layout(state):
    '''Create all the layout widgets for Xplore page

    This helps avoid bad screen effects such as screen shifting, 
    jerkiness etc. when changing or updating the page
    '''
    grid1 = st.container()
    with grid1:
        headercol, uniq_col = st.columns(2)
        header_ph = headercol.empty()
        uniq_ph = uniq_col.empty()

        scol1, scol2 = st.columns(2)
        scol1_ph = scol1.empty()
        scol2_ph = scol2.empty()
        if state.table in ['interfaces', 'ospf', 'bgp', 'evpnVni']:
            validate_expander = st.expander(
                'Assert', expanded=state.assert_clicked)
        else:
            validate_expander = None
        table_expander = st.expander('Table', expanded=True)

    return {
        'header_ph': header_ph,  # summary header
        'uniq_col_ph': uniq_ph,  # unique column selector
        'summ_ph': scol1_ph,  # summary table
        'uniq_ph': scol2_ph,  # unique table
        'assert_expander': validate_expander,  # assert expander
        'table_expander': table_expander,  # table dataframe expander
    }


def xplore_run():
    wsstate = st.session_state
    state = wsstate.xploreSessionState

    if not state.table:
        return

    layout = xplore_create_layout(state)
    sqobj = get_sqobject(state.table)
    df = gui_get_df(sqobj, _table=state.table,
                    namespace=state.namespace.split(),
                    hostname=state.hostname.split(),
                    start_time=state.start_time, end_time=state.end_time,
                    view=state.view, columns=state.columns)

    wsstate['xplore_df'] = df
    if not df.empty:
        if 'error' in df.columns:
            st.error(df.iloc[0].error)
            st.experimental_set_query_params(**asdict(state))
            st.stop()
            return

    if state.query:
        try:
            show_df = df.query(state.query).reset_index(drop=True)
            query_str = state.query
        except Exception:
            st.warning('Query string throws an exception, ignoring')
            show_df = df
            query_str = ''
    else:
        show_df = df
        query_str = ''

    if not show_df.empty:
        xplore_draw_summary_df(layout, state.table, state, sqobj, query_str)
        xplore_draw_uniq_histogram(layout, state.table, state,
                                   wsstate, show_df)
        xplore_draw_assert_df(layout, state, sqobj)

    xplore_draw_table_df(layout, state.table, show_df)

    st.experimental_set_query_params(**asdict(state))
    state.state_changed = False


def xplore_draw_summary_df(layout, table, state, sqobj, query_str):
    '''Display the summary dataframe'''
    summ_df = xplore_run_summarize(sqobj,
                                   namespace=state.namespace.split(),
                                   hostname=state.hostname.split(),
                                   start_time=state.start_time,
                                   end_time=state.end_time,
                                   query_str=query_str)

    header_col = layout['header_ph']
    header_col.write(
        f'<h2 style="color: darkblue; font-weight: bold;">{state.table} View</h2>',
        unsafe_allow_html=True)

    summ_ph = layout['summ_ph']
    if not summ_df.empty:
        summ_ph.subheader('Summary Information')
        summ_ph.dataframe(data=summ_df.astype(str))


def xplore_draw_uniq_histogram(layout, table, state, wsstate, show_df):
    '''Display the unique histogram'''

    dfcols = show_df.columns.tolist()
    if (state.table == 'routes' and 'prefix' in dfcols and
            'prefixlen' not in dfcols):
        dfcols.append('prefixlen')

    uniq_col = layout['uniq_col_ph']
    dfcols = sorted((filter(lambda x: x not in ['index', 'sqvers'],
                            dfcols)))
    with uniq_col:
        if (state.uniq_clicked == '-' or
                state.uniq_clicked not in dfcols):
            if 'hostname' in dfcols:
                selindex = dfcols.index('hostname')+1
                uniq_sel = 'hostname'
            else:
                uniq_sel = '-'
                selindex = 0
        elif state.uniq_clicked in dfcols:
            uniq_sel = state.uniq_clicked
            selindex = dfcols.index(state.uniq_clicked)+1
        else:
            uniq_sel = '-'
            selindex = 0

        if 'xplore_uniq_col' not in wsstate:
            state.uniq_clicked = st.selectbox(
                'Distribution Count of', options=['-'] + dfcols,
                index=selindex, key='xplore_uniq_col', on_change=xplore_sync_state)
        else:
            wsstate.xplore_uniq_col = uniq_sel
            state.uniq_clicked = st.selectbox(
                'Distribution Count of', options=['-'] + dfcols,
                index=selindex,
                key='xplore_uniq_col', on_change=xplore_sync_state)

    scol2 = layout['uniq_ph']

    if state.uniq_clicked != '-':
        uniq_df = xplore_run_unique(show_df,
                                    columns=state.uniq_clicked)
    else:
        uniq_df = pd.DataFrame()

    if not uniq_df.empty:
        if uniq_df.shape[0] > 16:
            scol2.warning(
                f'{state.uniq_clicked} has cardinality of {uniq_df.shape[0]}. Displaying top 16')
            chart = alt.Chart(
                uniq_df.head(16),
                title=f'{state.uniq_clicked} Distribution') \
                .mark_bar(color='purple', tooltip=True) \
                .encode(y=alt.Y(f'{state.uniq_clicked}:N',
                                sort='-x'),
                        x='numRows')
        else:
            chart = alt.Chart(
                uniq_df, title=f'{state.uniq_clicked} Distribution') \
                .mark_bar(color='purple', tooltip=True) \
                .encode(y=alt.Y(f'{state.uniq_clicked}:N',
                                sort='-x'),
                        x='numRows')
        scol2.altair_chart(chart)


def xplore_draw_assert_df(layout, state, sqobj):
    '''Display the assert dataframe'''
    if state.assert_clicked:
        assert_df = xplore_run_assert(sqobj,
                                      start_time=state.start_time,
                                      end_time=state.end_time,
                                      namespace=state.namespace.split())
    else:
        assert_df = pd.DataFrame()

    if state.table in ['interfaces', 'ospf', 'bgp', 'evpnVni']:
        if assert_df.empty:
            expand_assert = False
        else:
            expand_assert = True
        assert_expander = layout['assert_expander']
        with assert_expander:
            if not assert_df.empty:
                st.dataframe(data=assert_df)
            elif state.assert_clicked:
                st.write('Assert passed')
            else:
                st.write('Assert not run')


def xplore_draw_table_df(layout, table, show_df):
    '''Display the table dataframe'''
    expander = layout['table_expander']
    with expander:
        if show_df.shape[0] > 256:
            st.write(
                f'Showing first 256 of {show_df.shape[0]} rows, use query to filter')
        if not show_df.empty:
            convert_dict = {
                x: 'str' for x in show_df.select_dtypes('category').columns}
            st.dataframe(data=sq_gui_style(show_df.head(256)
                                           .astype(convert_dict),
                                           table),
                         height=600, width=2500)
        else:
            st.warning('No Data from query')



import re
from suzieq.sqobjects.basicobj import SqObject
from suzieq.utils import convert_macaddr_format_to_colon


class MacsObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='macs', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'macaddr',
                                'remoteVtepIp', 'vlan', 'localOnly', 'bd',
                                'moveCount', 'query_str']
        self._convert_args = {
            'macaddr': convert_macaddr_format_to_colon
        }

    def validate_get_input(self, **kwargs):
        for key in kwargs:
            if key == 'vlan':
                for ele in kwargs[key]:
                    if (ele.startswith(('<', '>', '!')) and (
                            ele in ['<', '<=', '>', '>=', '!=', '!'])):
                        raise ValueError('operator must not be separated by '
                                         'space, as in "<20"')
                    words = re.split(r'<|<=|>|>=|!', ele)
                    try:
                        int(words[-1])
                    except Exception:
                        raise ValueError(f'Invalid VLAN value: {kwargs[key]}')
        super().validate_get_input(**kwargs)



import numpy as np
from .engineobj import SqPandasEngine


class LldpObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'lldp'

    def get(self, **kwargs):
        '''A separate get to handle JunOS MACaddr as peer'''

        addnl_fields = kwargs.get('addnl_fields', [])
        namespace = kwargs.get('namespace', [])
        columns = kwargs.get('columns', [])
        fields = kwargs.get('fields', [])
        dropcols = []

        if columns == ['default']:
            needed_fields = ['subtype', 'peerMacaddr', 'peerIfindex']
        elif 'peerIfname' in columns:
            needed_fields = ['namespace', 'hostname', 'ifname', 'peerHostname',
                             'subtype', 'peerMacaddr', 'peerIfindex']
        else:
            needed_fields = []

        for f in needed_fields:
            if f not in columns:
                addnl_fields.append(f)
                dropcols.append(f)

        df = super().get(addnl_fields=addnl_fields, **kwargs)
        if df.empty or (not needed_fields and columns != ['*']):
            return df

        macdf = df.query('subtype.isin(["", "mac address"])')
        if not macdf.empty:
            macs = df.peerMacaddr.unique().tolist()
            addrdf = self._get_table_sqobj('address').get(
                namespace=namespace, address=macs,
                columns=['namespace', 'hostname', 'ifname', 'macaddr'])

            if not addrdf.empty:
                df = df.merge(
                    addrdf, how='left',
                    left_on=['namespace', 'peerHostname', 'peerMacaddr'],
                    right_on=['namespace', 'hostname', 'macaddr'],
                    suffixes=['', '_y']) \
                    .fillna({'ifname_y': '-'}) \

                if not df.empty:
                    df['peerIfname'] = np.where(df['peerIfname'] == '-',
                                                df['ifname_y'], df['peerIfname'])
                    df = df.drop(columns=['hostname_y', 'ifname_y',
                                          'timestamp_y', 'vrf', 'macaddr'])

        ifidx_df = df.query('subtype.str.startswith("locally")')
        ifindices = ifidx_df.query('peerIfindex != 0').peerIfindex \
            .unique().tolist()
        if not ifidx_df.empty and ifindices:
            ifdf = self._get_table_sqobj('interfaces').get(
                namespace=namespace, ifindex=ifindices,
                columns=['namespace', 'hostname', 'ifname', 'ifindex'])
            df = df.merge(
                ifdf, how='left',
                left_on=['namespace', 'peerHostname', 'peerIfindex'],
                right_on=['namespace', 'hostname', 'ifindex'],
                suffixes=['', '_y']) \
                .fillna({'ifname_y': '-'})

            if not df.empty:
                df['peerIfname'] = np.where(df['peerIfname'] == '-',
                                            df['ifname_y'], df['peerIfname'])

        dropcols.extend(['hostname_y', 'ifname_y', 'timestamp_y', 'ifindex'])
        return df.drop(columns=dropcols, errors='ignore')

    def summarize(self, **kwargs):
        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('nbrCnt', 'hostname', 'count'),
            ('peerHostnameCnt', 'peerHostname', 'count'),
            ('uniquePeerMgmtIPCnt', 'mgmtIP', 'nunique'),
        ]

        self._summarize_on_add_with_query = [
            ('missingPeerInfoCnt', "mgmtIP == '-' or mgmtIP == ''", 'mgmtIP')
        ]

        return super().summarize(**kwargs)



from importlib.util import find_spec
import os
import base64
from enum import Enum

import streamlit as st
import pandas as pd
from streamlit.elements.select_slider import SelectSliderMixin
from streamlit.server.server import Server
from streamlit.report_thread import get_report_ctx


class SuzieqMainPages(str, Enum):
    STATUS = "Status"
    XPLORE = "Xplore"
    PATH = "Path"
    SEARCH = "Search"


@st.cache(ttl=90, allow_output_mutation=True, show_spinner=False,
          max_entries=20)
def gui_get_df(sqobject, verb: str = 'get', **kwargs) -> pd.DataFrame:
    """Get the cached value of the table provided

    The only verbs supported are get and find.

    Args:
        sqobject ([type]): The sqobject for which to get the data
        verb (str, optional): . Defaults to 'get'.

    Returns:
        [pandas.DataFrame]: The dataframe

    Raises:
        ValueError: If the verb is not supported
    """
    table = kwargs.pop('_table', '')
    view = kwargs.pop('view', 'latest')
    columns = kwargs.pop('columns', ['default'])
    stime = kwargs.pop('start_time', '')
    etime = kwargs.pop('end_time', '')

    if columns == ['all']:
        columns = ['*']
    if verb == 'get':
        df = sqobject(view=view, start_time=stime, end_time=etime) \
            .get(columns=columns, **kwargs)
    elif verb == 'find':
        df = sqobject(view=view, start_time=stime, end_time=etime) \
            .find(**kwargs)
    else:
        raise ValueError(f'Unsupported verb {verb}')

    if not df.empty:
        df = sqobject().humanize_fields(df)
        if table == 'address':
            if 'ipAddressList' in df.columns:
                df = df.explode('ipAddressList').fillna('')
            if 'ip6AddressList' in df.columns:
                df = df.explode('ip6AddressList').fillna('')
    if not (columns == ['*'] or columns == ['default']):
        return df[columns].reset_index(drop=True)
    return df.reset_index(drop=True)


def get_base_url():
    '''Return the base URL of the page.
    Usually connections are http://localhost:8501. But the port can change
    or it can be a remote connection. And so, its useful to get the base
    URL for use with links on various pages.
    '''
    session_id = get_report_ctx().session_id
    session_info = Server.get_current()._get_session_info(session_id)

    if session_info:
        return f'http://{session_info.ws.request.host}/'

    return 'http://localhost:8501/'


def get_session_id():
    '''Return Streamlit's session ID'''
    return get_report_ctx().session_id


def get_main_session_by_id(session_id):
    """This returns the session state based on the ID provided

    This is relying on Streamlit internals and is not a public API.

    Args:
        session_id ([type]): The session id string

    Returns:
        [type]: session state associated with session or None
    """
    session = Server.get_current()._session_info_by_id.get(session_id, None)
    if session:
        return session.session.session_state

    return None


def get_image_dir():
    '''Get directory where images are stored'''
    return(os.path.dirname(find_spec('suzieq.gui')
                           .loader.path) + '/images')


def display_help_icon(url: str):
    '''Display Help Icon with click to take you to appropriate page'''
    help_img = f'{get_image_dir()}/helps.png'
    st.sidebar.markdown(f'<a target="_help" href="{url}"><img class="help-img" src="data:image/png;base64,{base64.b64encode(open(help_img, "rb").read()).decode()}"></a>',
                        unsafe_allow_html=True)


def maximize_browser_window():
    '''Maximize browser window in streamlit'''

    max_width_str = "max-width: 2000px;"
    st.markdown(
        f"""
    <style>
    .reportview-container .main .block-container{{
        {max_width_str}
    }}
    </style>
    """,
        unsafe_allow_html=True,
    )


def horizontal_radio():
    '''Make the radio buttons horizontal'''
    st.write('<style>div.row-widget.stRadio > div{flex-direction:row;}</style>',
             unsafe_allow_html=True)


def hide_st_index():
    '''CSS to hide table index rendered via st.table'''
    st.markdown("""
        <style>
        .table td:nth-child(1) {
            display: none;
        }
        .table th:nth-child(1) {
            display: none;
        }
        </style>
        """, unsafe_allow_html=True)


def color_row(row, **kwargs):
    """Color the appropriate column red if the status has failed"""
    fieldval = kwargs.pop("fieldval", "down")
    field = kwargs.pop("field", "state")
    color = kwargs.pop("color", "black")
    bgcolor = kwargs.pop("bgcolor", "yellow")

    if row[field] in fieldval:
        return [f"background-color: {bgcolor}; color: {color}"]*len(row)
    else:
        return [""]*len(row)


def color_element_red(value, **kwargs):
    '''Use with applymap to color a cell based on a value'''
    fieldval = kwargs.pop("fieldval", "down")
    if value not in fieldval:
        return "background-color: red; color: white;"


def ifstate_red(row):
    '''Color interface state red if admin state is up, but not oper state'''
    llen = len(row.index.tolist())
    if row.adminState == "up" and row.state == "down":
        return ["background-color: red; color: white;"]*llen
    else:
        return ["background-color: white; color: black;"]*llen


def sq_gui_style(df, table, is_assert=False):
    """Apply appropriate styling for the dataframe based on the table"""

    if is_assert:
        if not df.empty:
            return df.style.apply(color_row, axis=1, field='status',
                                  fieldval=['fail'], bgcolor='darkred',
                                  color='white')
        else:
            return df

    if table == 'bgp' and 'state' in df.columns:
        return df.style.hide_index() \
            .applymap(color_element_red, fieldval=['Established'],
                      subset=pd.IndexSlice[:, ['state']])
    elif table == 'ospf' and 'adjState' in df.columns:
        return df.style.hide_index() \
            .applymap(color_element_red,
                      fieldval=["full", "passive"],
                      subset=pd.IndexSlice[:, ['adjState']])

    elif table == "routes" and 'prefix' in df.columns:
        return df.style.hide_index() \
            .apply(color_row, axis=1, fieldval=['0.0.0.0/0'],
                   field='prefix')
    elif table == "interfaces" and 'state' in df.columns:
        return df.style.hide_index().apply(ifstate_red, axis=1)
    elif table == "device":
        return df.style.hide_index() \
            .apply(color_row, axis=1, fieldval=['dead', 'neverpoll'],
                   field='status', bgcolor='red', color='white')
    else:
        return df.style.hide_index()


def get_color_styles(color: str) -> str:
    """Compile some hacky CSS to override the theme color."""
    # fmt: off
    color_selectors = ["a", "a:hover", "*:not(textarea).st-ex:hover", ".st-en:hover"]
    bg_selectors = [".st-da", "*:not(button).st-en:hover"]
    border_selectors = [".st-ft", ".st-fs", ".st-fr", ".st-fq", ".st-ex:hover", ".st-en:hover"]
    # fmt: on
    css_root = "#root { --primary: %s }" % color
    css_color = ", ".join(color_selectors) + "{ color: %s !important }" % color
    css_bg = ", ".join(bg_selectors) + \
        "{ background-color: %s !important }" % color
    css_border = ", ".join(border_selectors) + \
        "{ border-color: %s !important }" % color
    other = ".decoration { background: %s !important }" % color
    return f"<style>{css_root}{css_color}{css_bg}{css_border}{other}</style>"


def display_title(page: str):
    '''Render the logo and the app name'''

    state = st.session_state
    pagelist = state.pagelist
    search_text = state.search_text

    LOGO_IMAGE = f'{get_image_dir()}/logo-small.jpg'
    st.markdown(
        """
        <style>
        .container {
            display: flex;
        }
        .logo-text {
            font-weight:700 !important;
            font-size:24px !important;
            color: purple !important;
            padding-top: 40px !important;
        }
        .logo-img {
            width: 20%;
            height: auto;
            float:right;
        }
        </style>
        """,
        unsafe_allow_html=True
    )

    title_col, mid, page_col, srch_col = st.columns([2, 1, 2, 2])
    with title_col:
        st.markdown(
            f"""
            <div class="container">
                <img class="logo-img" src="data:image/png;base64,{base64.b64encode(open(LOGO_IMAGE, "rb").read()).decode()}">
                <h1 style='color:purple;'>Suzieq</h1>
            </div>
            """,
            unsafe_allow_html=True
        )

    sel_pagelist = list(filter(lambda x: not x.startswith('_'), pagelist))

    with srch_col:
        st.text(' ')
        search_str = st.text_input("Search", value=state.search_text,
                                   key='search', on_change=main_sync_state)
    if search_text is not None and (search_str != search_text):
        # We're assuming here that the page is titled Search
        page = 'Search'

    with page_col:
        # The empty writes are for aligning the pages link with the logo
        st.text(' ')
        srch_holder = st.empty()
        pageidx = sel_pagelist.index(page or 'Status')
        if 'sq_page' not in state:
            page = srch_holder.selectbox('Page', sel_pagelist, index=pageidx,
                                         key='sq_page',
                                         on_change=main_sync_state)
        else:
            page = srch_holder.selectbox('Page', sel_pagelist, key='sq_page',
                                         on_change=main_sync_state)
            page = state.sq_page

    return page, search_str


def main_sync_state():

    wsstate = st.session_state

    if wsstate.search_text != wsstate.search:
        wsstate.search_text = wsstate.search
        wsstate.page = wsstate.sq_page = 'Search'

    if wsstate.page != wsstate.sq_page:
        wsstate.page = wsstate.sq_page
        st.experimental_set_query_params(**dict())
        if wsstate.page != 'Search':
            wsstate.search_text = ''



from suzieq.sqobjects.basicobj import SqObject
import pandas as pd
from suzieq.utils import humanize_timestamp


class DevconfigObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='devconfig', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns',
                                'query_str']



import time
from nubia import command, argument
import pandas as pd

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.fs import FsObj


@command("fs", help="Act on File System data")
class FsCmd(SqCommand):
    """Filesystem information such as total disk space, filesystems etc"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",
        query_str: str = " ",
        columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=FsObj,
        )

    @command("show")
    @argument("mountPoint", description="The mount point inside the FileSystem")
    @argument("used_percent", description="must be of the form "
              "[<|<=|>=|>|!]value. Eg: '<=20'")
    def show(self, mountPoint: str = '', used_percent: str = ''):
        """Show File System info
        """
        if self.columns is None:
            return

        # Get the default display field names
        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        if used_percent and not any(used_percent.startswith(x)
                                    for x in ['<=', '>=', '<', '>', '!']):
            try:
                int(used_percent)
            except ValueError:
                df = pd.DataFrame(
                    {'error': ['ERROR invalid used-percent operation']})
                return self._gen_output(df)

        df = self.sqobj.get(
            hostname=self.hostname,
            columns=self.columns,
            namespace=self.namespace,
            mountPoint=mountPoint.split(),
            usedPercent=used_percent,
            query_str=self.query_str,
        )
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)

    @command("summarize")
    def summarize(self):
        """
        Summarize the filesystem/storage info
        """
        df = pd.DataFrame({'error': ['Summarize not yet supported for FS']})
        return self._gen_output(df)



import numpy as np

from .engineobj import SqPandasEngine


class VlanObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'vlan'

    def get(self, **kwargs):
        """Get VLAN info based on specified keywords"""

        # ifname is a deprecated field, from version 1.0
        # vlanName is the correct fieldname. SO we need to do magic
        # to fix this. And thats why this routine exists

        dropcols = []
        addnl_fields = kwargs.pop('addnl_fields', [])
        columns = kwargs.pop('columns', [])
        if (columns != ['*'] and (columns == ['default'] or
                                  'ifname' not in columns)):
            if 'ifname' not in addnl_fields:
                addnl_fields.append('ifname')
                dropcols.append('ifname')

        df = super().get(addnl_fields=addnl_fields,
                         merge_fields={'ifname': 'vlanName'},
                         **kwargs)
        if not df.empty:
            df.drop(columns=dropcols, errors='ignore', inplace=True)

        return df

    def summarize(self, **kwargs):
        """Describe the IP Address data"""

        self._init_summarize(self.iobj.table, **kwargs)
        if self.summary_df.empty:
            return self.summary_df

        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('uniqueVlanCnt', 'vlan', 'nunique')
        ]

        self._summarize_on_add_with_query = [
            ('activeVlanCnt', 'state == "active"', 'vlan', 'nunique'),
            ('suspendedVlanCnt', 'state == "suspend"', 'vlan', 'nunique')
        ]

        self._summarize_on_perdevice_stat = [
            ('vlanPerDeviceStat', '', 'vlan', 'count')
        ]

        self._gen_summarize_data()

        self._add_stats_to_summary(self.summary_df.groupby(
            by=['namespace', 'vlan'])['interfaces'].count(), 'ifPerVlanStat',
            True)
        self.summary_row_order.append('ifPerVlanStat')

        self._post_summarize()
        return self.ns_df.convert_dtypes()



from suzieq.sqobjects.basicobj import SqObject


class VlanObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='vlan', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'vlan',
                                'state', 'vlanName', 'query_str']
        self._valid_arg_vals = {'state': ['active', 'suspended', '']}



from suzieq.sqobjects.basicobj import SqObject
import pandas as pd
from suzieq.utils import humanize_timestamp


class BgpObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='bgp', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'state',
                                'vrf', 'peer', 'asn', 'query_str']
        self._valid_arg_vals = {
            'state': ['Established', 'NotEstd', 'dynamic', ''],
            'status': ['all', 'pass', 'fail'],
        }
        self._valid_assert_args = ['namespace', 'hostname', 'vrf', 'status']

    def aver(self, **kwargs):
        """Assert that the BGP state is OK"""

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')
        try:
            self.validate_assert_input(**kwargs)
        except Exception as error:
            df = pd.DataFrame({'error': [f'{error}']})
            return df

        return self.engine.aver(**kwargs)

    def humanize_fields(self, df: pd.DataFrame, subset=None) -> pd.DataFrame:
        '''Humanize the timestamp and boot time fields'''
        if df.empty:
            return df

        if 'estdTime' in df.columns:
            df['estdTime'] = humanize_timestamp(df.estdTime,
                                                self.cfg.get('analyzer', {})
                                                .get('timezone', None))

        return df



from nubia import context
from nubia import exceptions
from nubia import eventbus
import sys

from suzieq.utils import load_sq_config, Schema
from suzieq.engines import get_sqengine


class NubiaSuzieqContext(context.Context):
    def __init__(self, engine="pandas"):
        self.cfg = None
        self.schemas = None

        self.pager = False
        self.namespace = ""
        self.hostname = ""
        self.start_time = ""
        self.end_time = ""
        self.exec_time = ""
        self.engine = engine
        self.col_width = 50
        self.sort_fields = []
        self.view = ""
        super().__init__()

    def on_connected(self, *args, **kwargs):
        if self._args.config:
            self.cfg = load_sq_config(validate=True,
                                      config_file=self._args.config)
        else:
            self.cfg = load_sq_config(validate=True)
        if not self.cfg:
            sys.exit(1)
        self.schemas = Schema(self.cfg["schema-directory"])

    def on_cli(self, cmd, args):
        # dispatch the on connected message
        self.verbose = args.verbose
        self.registry.dispatch_message(eventbus.Message.CONNECTED)

    def on_interactive(self, args):
        self.verbose = args.verbose
        ret = self._registry.find_command("connect").run_cli(args)
        if ret:
            raise exceptions.CommandError("Failed starting interactive mode")
        # dispatch the on connected message
        self.registry.dispatch_message(eventbus.Message.CONNECTED)

    def change_engine(self, engine: str):
        '''This doesn't work at this time'''
        if engine == self.engine:
            return

        self.engine = engine



import time
from nubia import command, argument

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.sqPoller import SqPollerObj


@command("sqPoller", help="Act on SqPoller data", aliases=['sqpoller'])
class SqPollerCmd(SqCommand):
    """Information about the poller"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",
        columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            sqobj=SqPollerObj,
        )

    @command("show")
    @argument("service", description="name of service to match")
    @argument("status", description="status of service to match",
              choices=["all", "pass", "fail"])
    def show(self, service: str = "", status: str = "all"):
        """Show Suzieq poller info such as status of polled commands etc.
        """
        if self.columns is None:
            return

        # Get the default display field names
        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        df = self.sqobj.get(
            hostname=self.hostname,
            columns=self.columns,
            service=service,
            status=status,
            namespace=self.namespace,
        )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)



import pandas as pd

from suzieq.engines.pandas.engineobj import SqPandasEngine


class TableObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'tables'

    def get(self, **kwargs):
        """Show the known tables for which we have information"""

        table_list = self._dbeng.get_tables()
        df = pd.DataFrame()
        columns = kwargs.pop('columns', ['default'])
        unknown_tables = []
        tables = []

        for table in table_list:
            table_obj = self._get_table_sqobj(table)

            if not table_obj:
                # This is a table without an sqobject backing store
                # this happens either because we haven't yet implemented the
                # table functions or because this table is collapsed into a
                # single table as in the case of ospf
                unknown_tables.append(table)
                table_inst = self._get_table_sqobj('tables')
                table_inst._table = table
            else:
                table_inst = table_obj

            info = {'table': table}
            info.update(table_inst.get_table_info(
                table, columns=['namespace', 'hostname', 'timestamp'],
                **kwargs))
            tables.append(info)

        df = pd.DataFrame.from_dict(tables)
        if df.empty:
            return df

        df = df.sort_values(by=['table']).reset_index(drop=True)
        cols = df.columns
        total = pd.DataFrame([['TOTAL',  df['firstTime'].min(),
                               df['latestTime'].max(),
                               df['intervals'].max(),
                               df['allRows'].sum(),
                               df['namespaces'].max(),
                               df['deviceCnt'].max()]],
                             columns=cols)
        df = df.append(total, ignore_index=True).dropna()
        return df

    def summarize(self, **kwargs):
        df = self.get(**kwargs)

        if df.empty or ('error' in df.columns):
            return df

        df = df.set_index(['table'])

        sdf = pd.DataFrame({
            'serviceCnt': [df.index.nunique()-1],
            'namespaceCnt': [df.at['TOTAL', 'namespaces']],
            'deviceCnt': [df.at['device', 'deviceCnt']],
            'earliestTimestamp': [df.firstTime.min()],
            'lastTimestamp': [df.latestTime.max()],
            'firstTime99': [df.firstTime.quantile(0.99)],
            'latestTime99': [df.latestTime.quantile(0.99)],
        })
        return sdf.T.rename(columns={0: 'summary'})

    def top(self, **kwargs):
        "Tables implementation of top has to eliminate the TOTAL row"

        what = kwargs.pop("what", None)
        reverse = kwargs.pop("reverse", False)
        sqTopCount = kwargs.pop("count", 5)

        if not what:
            return pd.DataFrame()

        df = self.get(addnl_fields=self.iobj._addnl_fields, **kwargs)
        if df.empty or ('error' in df.columns):
            return df

        if reverse:
            return df.query('table != "TOTAL"') \
                .nsmallest(sqTopCount, columns=what, keep="all") \
                .head(sqTopCount)
        else:
            return df.query('table != "TOTAL"') \
                .nlargest(sqTopCount, columns=what, keep="all") \
                .head(sqTopCount)



from suzieq.sqobjects.basicobj import SqObject
import pandas as pd


class RoutesObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='routes', **kwargs)
        self._addnl_filter = 'metric != 4278198272'
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'prefix',
                                'vrf', 'protocol', 'prefixlen', 'ipvers',
                                'add_filter', 'address', 'query_str']

    def lpm(self, **kwargs):
        '''Get the lpm for the given address'''
        if not kwargs.get("address", None):
            raise AttributeError('ip address is mandatory parameter')
        if isinstance(kwargs['address'], list):
            if len(kwargs['address']) > 1:
                raise AttributeError('Just one address at a time')
            kwargs['address'] = kwargs['address'][0]
        return self.engine.lpm(**kwargs)

    def summarize(self, namespace=[], vrf=[], hostname=[], query_str=''):
        """Summarize routing info for one or more namespaces"""
        if self.columns != ["default"]:
            self.summarize_df = pd.DataFrame(
                {'error':
                 ['ERROR: You cannot specify columns with summarize']})
            return self.summarize_df
        return self.engine.summarize(namespace=namespace, vrf=vrf,
                                     query_str=query_str,
                                     hostname=hostname)



from suzieq.sqobjects.basicobj import SqObject


class OspfIfObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='ospf', **kwargs)
        self._valid_get_args = ['namespace',
                                'hostname', 'columns', 'ifname', 'query_str']



from ipaddress import ip_network
import pandas as pd
import numpy as np

from suzieq.exceptions import NoLLdpError
from .engineobj import SqPandasEngine


class InterfacesObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'interfaces'

    def get(self, **kwargs):
        """Handling state outside of regular filters"""
        state = kwargs.pop('state', '')
        iftype = kwargs.pop('type', '')
        ifname = kwargs.get('ifname', '')
        vrf = kwargs.pop('vrf', '')
        master = kwargs.pop('master', [])

        if vrf:
            master.extend(vrf)

        if iftype and iftype != ["all"]:
            df = super().get(type=iftype, master=master, **kwargs)
        else:
            df = super().get(master=master, **kwargs)

        if df.empty:
            return df

        if state:
            df = df.query(f'state=="{state}"')

        if not (iftype or ifname) and 'type' in df.columns:
            return df.query('type != "internal"').reset_index(drop=True)
        else:
            return df.reset_index(drop=True)

    def aver(self, what="", **kwargs) -> pd.DataFrame:
        """Assert that interfaces are in good state"""

        if what == "mtu-value":
            result_df = self._assert_mtu_value(**kwargs)
        else:
            result_df = self._assert_interfaces(**kwargs)
        return result_df

    def summarize(self, **kwargs) -> pd.DataFrame:
        """Summarize interface information"""
        self._init_summarize(self.iobj._table, **kwargs)
        if self.summary_df.empty:
            return self.summary_df

        # Loopback interfaces on Linux have "unknown" as state
        self.summary_df["state"] = self.summary_df['state'] \
                                       .map({"unknown": "up",
                                             "up": "up", "down": "down",
                                             "notConnected": "notConnected"})

        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('interfaceCnt', 'ifname', 'count'),
        ]

        self._summarize_on_add_with_query = [
            ('devicesWithL2Cnt', 'master == "bridge"', 'hostname', 'nunique'),
            ('devicesWithVxlanCnt', 'type == "vxlan"', 'hostname'),
            ('ifDownCnt', 'state != "up" and adminState == "up"', 'ifname'),
            ('ifAdminDownCnt', 'adminState != "up"', 'ifname'),
            ('ifWithMultipleIPCnt', 'ipAddressList.str.len() > 1', 'ifname'),
        ]

        self._summarize_on_add_list_or_count = [
            ('uniqueMTUCnt', 'mtu'),
            ('uniqueIfTypesCnt', 'type'),
            ('speedCnt', 'speed'),
        ]

        self._summarize_on_add_stat = [
            ('ifChangesStat', 'type != "bond"', 'numChanges'),
        ]

        self._summarize_on_perdevice_stat = [
            ('ifPerDeviceStat', '', 'ifname', 'count')
        ]

        self._gen_summarize_data()

        # The rest of the summary generation is too specific to interfaces
        original_summary_df = self.summary_df
        self.summary_df = original_summary_df.explode(
            'ipAddressList').dropna(how='any')

        if not self.summary_df.empty:
            self.nsgrp = self.summary_df.groupby(by=["namespace"])
            self._add_field_to_summary(
                'ipAddressList', 'nunique', 'uniqueIPv4AddrCnt')
        else:
            self._add_constant_to_summary('uniqueIPv4AddrCnt', 0)
        self.summary_row_order.append('uniqueIPv4AddrCnt')

        self.summary_df = original_summary_df \
            .explode('ip6AddressList') \
            .dropna(how='any') \
            .query('~ip6AddressList.str.startswith("fe80:")')

        if not self.summary_df.empty:
            self.nsgrp = self.summary_df.groupby(by=["namespace"])
            self._add_field_to_summary(
                'ip6AddressList', 'nunique', 'uniqueIPv6AddrCnt')
        else:
            self._add_constant_to_summary('uniqueIPv6AddrCnt', 0)
        self.summary_row_order.append('uniqueIPv6AddrCnt')

        self._post_summarize(check_empty_col='interfaceCnt')
        return self.ns_df.convert_dtypes()

    def _assert_mtu_value(self, **kwargs) -> pd.DataFrame:
        """Workhorse routine to match MTU value"""

        columns = ["namespace", "hostname", "ifname", "state", "mtu",
                   "timestamp"]

        matchval = kwargs.pop('matchval', [])
        status = kwargs.pop('status', '')

        matchval = [int(x) for x in matchval]

        result_df = self.get(columns=columns, **kwargs) \
                        .query('ifname != "lo"')

        if not result_df.empty:
            result_df['status'] = result_df.apply(
                lambda x, matchval: 'pass' if x['mtu'] in matchval else 'fail',
                axis=1, args=(matchval,))

        if status == "fail":
            result_df = result_df.query('status == "fail"')
        elif status == "pass":
            result_df = result_df.query('status == "pass"')

        if not result_df.empty:
            return result_df.rename(columns={'status': 'assert'})
        else:
            return result_df

    def _assert_interfaces(self, **kwargs) -> pd.DataFrame:
        """Workhorse routine that validates MTU match for specified input"""
        columns = kwargs.pop('columns', [])
        status = kwargs.pop('status', 'all')
        stime = kwargs.pop('start_time', '')
        etime = kwargs.pop('end_time', '')
        ignore_missing_peer = kwargs.pop('ignore_missing_peer', False)

        def _check_field(x, fld1, fld2, reason):
            if x.skipIfCheck or x.indexPeer < 0:
                return []

            if x[fld1] == x[fld2]:
                return []
            return reason

        def _check_ipaddr(x, fld1, fld2, reason):
            # If we have no peer, don't check
            if x.skipIfCheck or x.indexPeer < 0:
                return []

            if len(x[fld1]) != len(x[fld2]):
                return reason

            if (len(x[fld1]) != 0):
                if (x[fld1][0].split('/')[1] == "32" or
                    (ip_network(x[fld1][0], strict=False) ==
                        ip_network(x[fld2][0], strict=False))):
                    return []
            else:
                return []

            return reason

        columns = ["namespace", "hostname", "ifname", "state", "type", "mtu",
                   "vlan", "adminState", "ipAddressList", "ip6AddressList",
                   "speed", "master", "timestamp", "reason"]

        if_df = self.get(columns=columns,
                         type=['ethernet', 'bond_slave', 'subinterface',
                               'vlan', 'bond'],
                         state='up',
                         **kwargs)
        if if_df.empty:
            if status != 'pass':
                if_df['assert'] = 'fail'
                if_df['assertReason'] = 'No data'

            return if_df

        # Map subinterface into parent interface
        if_df['pifname'] = if_df.apply(
            lambda x: x['ifname'].split('.')[0]
            if (x.type == 'subinterface') or (x.type == 'vlan')
            else x['ifname'], axis=1)

        # Thanks for Junos, remove all the useless parent interfaces if we have a
        # .0 interface since thats the real deal
        del_iflist = if_df.apply(lambda x: x.pifname
                                 if x['ifname'].endswith('.0') else '',
                                 axis=1) \
            .unique().tolist()

        if_df['type'] = if_df.apply(lambda x: 'ethernet'
                                    if x['ifname'].endswith('.0')
                                    else x['type'], axis=1)

        if_df = if_df.query(f'~ifname.isin({del_iflist})').reset_index()
        if_df['ifname'] = if_df.apply(
            lambda x: x['ifname'] if not x['ifname'].endswith('.0')
            else x['pifname'], axis=1)

        lldpobj = self._get_table_sqobj('lldp')
        vlanobj = self._get_table_sqobj('vlan')
        mlagobj = self._get_table_sqobj('mlag')

        # can't pass all kwargs, because lldp acceptable arguements are
        # different than interface
        namespace = kwargs.get('namespace', None)
        hostname = kwargs.get('hostname', None)
        lldp_df = lldpobj.get(namespace=namespace, hostname=hostname) \
                         .query('peerIfname != "-"')

        mlag_df = mlagobj.get(namespace=namespace, hostname=hostname)
        if not mlag_df.empty:
            mlag_peerlinks = set(mlag_df
                                 .groupby(by=['namespace', 'hostname', 'peerLink'])
                                 .groups.keys())
        else:
            mlag_peerlinks = set()

        # Get the VLAN info to ensure trunking ports are identical on both ends
        vlan_df = vlanobj.get(namespace=namespace, hostname=hostname)
        if not vlan_df.empty and not (vlan_df.interfaces.str.len() == 0).all():
            vlan_df = vlan_df.explode('interfaces') \
                             .drop(columns=['vlanName'], errors='ignore') \
                             .rename(columns={'interfaces': 'ifname',
                                              'vlan': 'vlanList'})

            vlan_df = vlan_df.groupby(by=['namespace', 'hostname', 'ifname'])[
                'vlanList'].unique().reset_index().query('ifname != ""')

            # OK merge this list into the interface info
            if_df = if_df.merge(vlan_df,
                                on=['namespace', 'hostname', 'ifname'],
                                how='left')
            if not if_df.empty:
                if_df['vlanList'] = if_df['vlanList'] \
                    .fillna({i: [] for i in if_df.index})

        if 'vlanList' not in if_df.columns:
            if_df['vlanList'] = []

        if lldp_df.empty:
            if status != 'pass':
                if_df['assertReason'] = 'No LLDP peering info'
                if_df['assert'] = 'fail'

            return if_df

        # Now create a single DF where you get the MTU for the lldp
        # combo of (namespace, hostname, ifname) and the MTU for
        # the combo of (namespace, peerHostname, peerIfname) and then
        # pare down the result to the rows where the two MTUs don't match
        idf = (
            pd.merge(
                if_df,
                lldp_df,
                left_on=["namespace", "hostname", "pifname"],
                right_on=['namespace', 'hostname', 'ifname'],
                how="outer",
            )
            .drop(columns=['ifname_y', 'timestamp_y'])
            .rename({'ifname_x': 'ifname', 'timestamp_x': 'timestamp'}, axis=1)
        )
        idf_nonsubif = idf.query('~type.isin(["subinterface", "vlan"])')
        idf_subif = idf.query('type.isin(["subinterface", "vlan"])')

        # Replace the bond_slave port interface with the bond interface

        idf_nonsubif = idf_nonsubif.merge(
            idf_nonsubif,
            left_on=["namespace", "peerHostname", "peerIfname"],
            right_on=['namespace', 'hostname', 'ifname'],
            how="outer", suffixes=["", "Peer"])

        idf_subif = idf_subif.merge(
            idf_subif,
            left_on=["namespace", "peerHostname", "peerIfname", 'vlan'],
            right_on=['namespace', 'hostname', 'pifname', 'vlan'],
            how="outer", suffixes=["", "Peer"])

        combined_df = pd.concat(
            [idf_subif, idf_nonsubif]).reset_index(drop=True)

        combined_df = combined_df \
            .drop(columns=["hostnamePeer", "pifnamePeer",
                           "mgmtIP", "description"]) \
            .dropna(subset=['hostname', 'ifname']) \
            .drop_duplicates(subset=['namespace', 'hostname', 'ifname'])

        if combined_df.empty:
            if status != 'pass':
                if_df['assertReason'] = 'No LLDP peering info'
                if_df['assert'] = 'fail'

            return if_df

        combined_df = combined_df.fillna(
            {'mtuPeer': 0, 'speedPeer': 0, 'typePeer': '',
             'peerHostname': '', 'peerIfname': '', 'indexPeer': -1})
        for fld in ['ipAddressListPeer', 'ip6AddressListPeer', 'vlanListPeer']:
            combined_df[fld] = combined_df[fld] \
                .apply(lambda x: x if isinstance(x, np.ndarray) else [])

        combined_df['assertReason'] = combined_df.apply(
            lambda x: []
            if (x['adminState'] == 'down' or
                (x['adminState'] == "up" and x['state'] == "up"))
            else [x.reason or "Interface Down"], axis=1)

        known_hosts = set(combined_df.groupby(by=['namespace', 'hostname'])
                          .groups.keys())
        # Mark interfaces that can be skippedfrom checking because you cannot find a
        # peer
        combined_df['skipIfCheck'] = combined_df.apply(
            lambda x:
            True if ((x.master == 'bridge') or
                     (x.type in ['bond_slave', 'vlan'])) else False,
            axis=1)

        combined_df['indexPeer'] = combined_df.apply(
            lambda x, kh: x.indexPeer
            if (x.namespace, x.hostname) in kh else -2,
            args=(known_hosts,), axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x: ['No Peer Found']
            if x.indexPeer == -1 and not x.skipIfCheck else [],
            axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x: ['Unpolled Peer']
            if x.indexPeer == -2 and not x.skipIfCheck else [],
            axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x: _check_field(x, 'mtu', 'mtuPeer', ['MTU mismatch']),
            axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x: _check_field(
                x, 'speed', 'speedPeer', ['Speed mismatch']),
            axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x: []
            if (x.indexPeer < 0 or
                ((x['type'] == x['typePeer']) or
                 (x['type'] == 'vlan' and x['typePeer'] == 'subinterface') or
                    (x['type'].startswith('ether') and
                     x['typePeer'].startswith('ether'))))
            else ['type mismatch'],
            axis=1)

        combined_df['assertReason'] += combined_df.apply(
            lambda x:
            _check_ipaddr(x, 'ipAddressList', 'ipAddressListPeer',
                          ['IP address mismatch']), axis=1)

        # We ignore MLAG peerlinks mainly because of NXOS erroneous output.
        # NXOS displays the VLANs associated with an interface via show vlan
        # which is then further pruned out by vPC. This pruned out list needs
        # to be extracted from the vPC output and used for the peerlink
        # instead of the output of show vlan for that interface. Since most
        # platforms perform their own MLAG consistency checks, we can skip
        # doing VLAN consistency check on the peerlink.
        # TODO: A better checker for MLAG peerlinks if needed at a later time.

        combined_df['assertReason'] += combined_df.apply(
            lambda x, mlag_peerlinks: []
            if ((x.indexPeer > 0 and
                ((x.namespace, x.hostname, x.master) not in mlag_peerlinks) and
                 (set(x['vlanList']) == set(x['vlanListPeer']))) or
                ((x.indexPeer < 0) or ((x.namespace, x.hostname, x.master) in mlag_peerlinks)))
            else ['VLAN set mismatch'], args=(mlag_peerlinks,), axis=1)

        if ignore_missing_peer:
            combined_df['check'] = combined_df.apply(
                lambda x: 'fail'
                if (len(x.assertReason) and
                    (x.assertReason[0] != 'No Peer Found'))
                else 'pass', axis=1)
        else:
            combined_df['check'] = combined_df.apply(
                lambda x: 'fail' if (len(x.assertReason)) else 'pass', axis=1)

        if status == "fail":
            combined_df = combined_df.query('check == "fail"').reset_index()
        elif status == "pass":
            combined_df = combined_df.query('check == "pass"').reset_index()

        return combined_df[['namespace', 'hostname', 'ifname', 'state',
                            'peerHostname', 'peerIfname', 'check',
                            'assertReason', 'timestamp']] \
            .rename({'check': 'assert'}, axis=1)



import numpy as np
import pandas as pd
import ipaddress

from .engineobj import SqPandasEngine


class EvpnvniObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'evpnVni'

    def get(self, **kwargs) -> pd.DataFrame:
        """Class-specific to extract info from addnl tables"""

        drop_cols = []
        addnl_fields = kwargs.pop('addnl_fields', [])
        columns = kwargs.get('columns', [])
        if (columns != ['*'] and 'ifname' not in columns and
                'ifname' not in addnl_fields):
            addnl_fields.append('ifname')
            drop_cols.append('ifname')

        if (columns == ['default'] or columns == ['*'] or
                'remoteVtepCnt' in columns):
            getVtepCnt = True
            if columns != ['*'] and 'remoteVtepList' not in columns:
                addnl_fields.append('remoteVtepList')
                drop_cols.append('remoteVtepList')
        else:
            getVtepCnt = False

        df = super().get(addnl_fields=addnl_fields, **kwargs)
        if df.empty:
            return df

        if getVtepCnt:
            insidx = df.columns.tolist().index('remoteVtepList')
            df.insert(insidx+1, 'remoteVtepCnt',
                      df.remoteVtepList.str.len())

        # See if we can retrieve the info to fill out the rest of the data
        # Start with VLAN values
        if 'vlan' not in df.columns and 'vrf' not in df.columns:
            return df.drop(columns=drop_cols, errors='ignore')

        iflist = list(set([x for x in df[df.vlan == 0]['ifname'].to_list()
                           if x and x != 'None']))
        if iflist:
            ifdf = self._get_table_sqobj('interfaces').get(
                namespace=kwargs.get('namespace', []), ifname=iflist,
                columns=['namespace', 'hostname', 'ifname', 'state', 'vlan',
                         'vni'])

            if not ifdf.empty:
                df = df.merge(ifdf, on=['namespace', 'hostname', 'ifname'],
                              how='left', suffixes=('', '_y')) \
                    .drop(columns=['timestamp_y', 'state_y', 'vni_y'])

                df['vlan'] = np.where(
                    df['vlan_y'].notnull(), df['vlan_y'], df['vlan'])
                df = df.drop(columns=['vlan_y']) \
                       .fillna({'vlan': 0}) \
                       .astype({'vlan': int})

        # Now we try to see if we can get the VRF info for all L2 VNIs as well
        ifdf = self._get_table_sqobj('interfaces') \
            .get(namespace=kwargs.get('namespace', []), type=['vlan'],
                 columns=['namespace', 'hostname', 'vlan', 'master']) \
            .query('master != ""') \
            .reset_index(drop=True)
        if not ifdf.empty and not df.empty and 'vrf' in df.columns:
            df = df.merge(ifdf, left_on=['namespace', 'hostname', 'vlan'],
                          right_on=['namespace', 'hostname', 'vlan'],
                          how='outer', suffixes=('', '_y'))

            # We fix the VRF from L2 VNIs from the SVI interface's master info
            # and we drop those entries that are just local VLANs, without VNI
            df['vrf'] = np.where(df['vrf'] == "", df['master'], df['vrf'])
            df = df.drop(columns=['master', 'timestamp_y'], errors='ignore') \
                   .dropna(subset=['vni'])

        # Fill out the numMacs and numArps columns if we can
        if 'numMacs' in df.columns:
            vlanlist = list(set(df[df.numMacs == 0]['vlan'].to_list()))
        else:
            vlanlist = []
        if vlanlist:
            if 'numMacs' in df.columns:
                macdf = self._get_table_sqobj('macs').get(
                    namespace=kwargs.get('namespace'))
                df['numMacs'] = df.apply(self._count_macs, axis=1,
                                         args=(macdf, ))
        return df.drop(columns=drop_cols, errors='ignore')

    def _count_macs(self, x, df):
        return df[(df.namespace == x['namespace']) & (df.vlan == x['vlan']) &
                  (df.hostname == x['hostname'])].count()

    def summarize(self, **kwargs):
        self._init_summarize(self.iobj._table, **kwargs)
        if self.summary_df.empty:
            return self.summary_df

        self._summarize_on_add_field = [
            ('uniqueVtepCnt', 'priVtepIp', 'nunique'),
            ('uniqueVniCnt', 'vni', 'nunique'),
        ]

        l3vni_count = self.summary_df.query('type == "L3" and vni != 0') \
                                     .groupby(by=['namespace'])['vni'].count()
        for ns in self.ns.keys():
            if l3vni_count.get(ns, 0):
                self.ns[ns]['mode'] = 'symmetric'
            else:
                self.ns[ns]['mode'] = 'asymmetric'
        self.summary_row_order.append('mode')

        self._summarize_on_add_with_query = [
            ('uniqueL3VniCnt', 'type == "L3" and vni != 0', 'vrf', 'nunique'),
            ('uniqueL2VniCnt', 'type == "L2"', 'vni', 'nunique'),
            ('uniqueMulticastGroups', 'mcastGroup != "0.0.0.0"', 'mcastGroup',
             'nunique'),
            ('vnisUsingMulticast',
             'type == "L2" and replicationType == "multicast"',
             'vni', 'nunique'),
            ('vnisUsingIngressRepl',
             'type == "L2" and replicationType == "ingressBGP"',
             'vni', 'nunique')
        ]

        self._gen_summarize_data()

        # To summarize accurately, we need to explode the remoteVteps
        # column from a list to an individual entry for each
        # remoteVteps in that list. The resulting set can be huge if them
        # number of Vteps times the ports is huge.
        #
        # the 'explode' only works post pandas 0.25

        self.summary_df = self.summary_df.explode(
            'remoteVtepList').dropna(how='any')
        self.nsgrp = self.summary_df.groupby(by=["namespace"])

        if not self.summary_df.empty:
            herPerVtepCnt = self.summary_df.groupby(
                by=['namespace', 'hostname'])['remoteVtepList'].nunique()
            self._add_stats_to_summary(herPerVtepCnt, 'remoteVtepsPerVtepStat',
                                       filter_by_ns=True)
        self.summary_row_order.append('remoteVtepsPerVtepStat')

        self._post_summarize(check_empty_col='uniqueVtepCnt')
        return self.ns_df.convert_dtypes()

    def aver(self, **kwargs) -> pd.DataFrame:
        """Assert for EVPN Data"""

        assert_cols = ["namespace", "hostname", "vni", "vlan", "remoteVtepList", "vrf",
                       "mcastGroup", "type", "priVtepIp", "state", "l2VniList",
                       "ifname", "secVtepIp"]

        kwargs.pop("columns", None)  # Loose whatever's passed
        status = kwargs.pop('status', 'all')

        df = self.get(columns=assert_cols, **kwargs)
        if df.empty:
            df = pd.DataFrame(columns=assert_cols)
            if status != 'pass':
                df['assertReason'] = 'No data found'
                df['assert'] = 'fail'
            return df

        df["assertReason"] = [[] for _ in range(len(df))]

        # Gather the unique set of VTEPs per VNI
        vteps_df = df.explode(column='remoteVtepList') \
            .dropna(how='any') \
            .groupby(by=['vni', 'type'])['remoteVtepList'] \
            .aggregate(lambda x: x.unique().tolist()) \
            .reset_index() \
            .dropna(how='any') \
            .rename(columns={'remoteVtepList': 'allVteps'})

        if not vteps_df.empty:
            her_df = df.merge(vteps_df)
        else:
            her_df = pd.DataFrame()

        if (not her_df.empty and
                (her_df.remoteVtepList.str.len() != 0).any()):
            # Check if every VTEP we know is reachable
            rdf = self._get_table_sqobj('routes').get(
                namespace=kwargs.get('namespace'), vrf='default')
            if not rdf.empty:
                rdf['prefixlen'] = rdf['prefix'].str.split('/') \
                                                    .str[1].astype('int')
            self._routes_df = rdf

            her_df["assertReason"] += her_df.apply(
                self._is_vtep_reachable, axis=1)

        mcast_df = df.query('mcastGroup != "0.0.0.0"')
        if not mcast_df.empty:
            # Ensure that all VNIs have at most one multicast group associated
            # per namespace

            mismatched_vni_df = mcast_df \
                .groupby(by=['namespace', 'vni'])['mcastGroup'] \
                .unique() \
                .reset_index() \
                .dropna() \
                .query('mcastGroup.str.len() != 1')

            if not mismatched_vni_df.empty:
                df['assertReason'] += df.apply(
                    lambda x, err_df: ['VNI has multiple mcast group']
                    if (x['namespace'] in err_df['namespace'] and
                        x['vni'] in err_df['vni']) else [], axis=1,
                    args=(mismatched_vni_df, ))
        elif not her_df.empty:
            # Every VTEP has info about every other VTEP for a given VNI
            her_df["assertReason"] += her_df.apply(
                self._all_vteps_present, axis=1)

        # State is up
        df["assertReason"] += df.apply(
            lambda x: ['interface is down']
            if x['type'] == "L2" and x['state'] != "up"
            else [], axis=1)

        devices = df["hostname"].unique().tolist()
        ifdf = self._get_table_sqobj('interfaces') \
            .get(namespace=kwargs.get("namespace", ""), hostname=devices,
                 type='vxlan')

        df = df.merge(ifdf[['namespace', 'hostname', 'ifname', 'master',
                            'vlan']],
                      on=['namespace', 'hostname', 'ifname'], how='left')

        # vxlan interfaces, if defined, for every VNI is part of bridge
        # We ensure this is true artificially for NXOS, ignored for JunOS.
        df["assertReason"] += df.apply(
            lambda x: ['vni not in bridge']
            if (x['type'] == "L2" and x['ifname'] != '-'
                and x['master'] != "bridge") else [],
            axis=1)

        # mac_df = MacsObj(context=self.ctxt) \
        #     .get(namespace=kwargs.get("namespace", ""),
        #          macaddr=["00:00:00:00:00:00"])

        # # Assert that we have HER for every remote VTEP
        # df['assertReason'] += df.apply(self._is_her_good,
        #                                args=(mac_df, ), axis=1)

        # Fill out the assert column
        df['assert'] = df.apply(lambda x: 'pass'
                                if not len(x.assertReason) else 'fail',
                                axis=1)

        if status == 'fail':
            df = df.query('assertReason.str.len() != 0')
        elif status == "pass":
            df = df.query('assertReason.str.len() == 0')

        return df[['namespace', 'hostname', 'vni', 'type',
                   'assertReason', 'assert', 'timestamp']] \
            .explode(column='assertReason') \
            .fillna({'assertReason': '-'})

    def _all_vteps_present(self, row):
        if row['secVtepIp'] == '0.0.0.0':
            myvteps = set([row['priVtepIp']])
        else:
            myvteps = set([row['secVtepIp'], row['priVtepIp']])

        isitme = set(row['allVteps']).difference(set(row['remoteVtepList']))
        if (isitme.intersection(myvteps) or
                (row['type'] == "L3" and row['remoteVtepList'] == ["-"])):
            return []

        return (['some remote VTEPs missing'])

    def _is_vtep_reachable(self, row):
        reason = []
        defrt = ipaddress.IPv4Network("0.0.0.0/0")
        for vtep in row['remoteVtepList'].tolist():
            if vtep == '-':
                continue

            # Have to hardcode lpm query here sadly to avoid reading of the
            # data repeatedly. The time for asserting the state of 500 VNIs
            # came down from 194s to 4s with the below piece of code instead
            # of invoking route's lpm to accomplish the task.
            cached_df = self._routes_df \
                            .query(f'namespace=="{row.namespace}" and hostname=="{row.hostname}"')
            route = self._get_table_sqobj('routes').lpm(vrf='default',
                                                        address=vtep,
                                                        cached_df=cached_df)

            if route.empty:
                reason += [f"{vtep} not reachable"]
                continue
            if route.prefix.tolist() == [defrt]:
                reason += [f"{vtep} reachable via default"]

        return reason

    def _is_her_good(self, row, mac_df):
        reason = []

        if row['type'] == 'L3':
            return reason

        her_list = mac_df[(mac_df['hostname'] == row['hostname']) &
                          (mac_df['namespace'] == row['namespace']) &
                          (mac_df['oif'] == row['ifname'])]['remoteVtepIp'] \
            .tolist()
        missing_hers = set(row['remoteVtepList']).difference(set(her_list))
        if not missing_hers:
            return reason
        return [f'HER is missing VTEPs {missing_hers}']



import pandas as pd

from suzieq.sqobjects.basicobj import SqObject
from suzieq.utils import humanize_timestamp


class IfObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='interfaces', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'ifname', 'columns',
                                'state', 'type', 'mtu', 'master', 'ifindex',
                                'vrf', 'query_str']
        self._valid_assert_args = ['namespace', 'hostname', 'ifname', 'what',
                                   'matchval', 'status', 'ignore_missing_peer']
        self._valid_arg_vals = {
            'state': ['up', 'down', 'notConnected', ''],
            'status': ['all', 'pass', 'fail'],
        }

    def aver(self, what='', **kwargs) -> pd.DataFrame:
        """Assert that interfaces are in good state"""
        try:
            self.validate_assert_input(**kwargs)
        except Exception as error:
            df = pd.DataFrame({'error': [f'{error}']})
            return df

        return self.engine.aver(what=what, **kwargs)

    def humanize_fields(self, df: pd.DataFrame, subset=None) -> pd.DataFrame:
        '''Humanize the timestamp and boot time fields'''
        if df.empty:
            return df

        if 'statusChangeTimestamp' in df.columns:
            df['statusChangeTimestamp'] = humanize_timestamp(
                df.statusChangeTimestamp,
                self.cfg.get('analyzer', {}).get('timezone', None))

        return df



import os
import pandas as pd
import pyarrow.parquet as pq
import pyarrow.dataset as ds    # put this later due to some numpy dependency
from datetime import datetime, timedelta, timezone
import logging
from typing import List
import tarfile
from itertools import repeat

from suzieq.utils import humanize_timestamp, SchemaForTable
from .migratedb import get_migrate_fn


class SqCoalesceState(object):
    '''Class that coalesces parquet files'''

    def __init__(self, logger: str, period: timedelta):
        self.current_df = pd.DataFrame()
        self.logger = logger if logger else logging.getLogger()
        self.keys = None        # filled as we tackle each table
        self.schema = None      # filled as we tackle each table
        self.dbeng = None       # filled as we tackle each table
        self.period = period
        self.prefix = 'sqc-h-'   # sqc == suzieq coalesceer, h: hourly coalesce
        self.ign_pfx = ['.', '_', 'sqc-']  # Prefixes to ignore for coalesceing
        self.wrfile_count = 0
        self.wrrec_count = 0
        self.poller_periods = set()
        self.block_start = self.block_end = 0

    def pq_file_name(self, *args):
        """Callback to create a filename that uses the timestamp of start
        of hour. This makes it easy for us to lookup data when we need to.

        :returns: parquet filename to write
        :rtype: str

        """
        # Using timestamp rather than date/time string to simplify reads
        return f'{self.prefix}{self.block_start}-{self.block_end}.parquet'


def archive_coalesced_files(filelist: List[str], outfolder: str,
                            state: SqCoalesceState, dodel: bool) -> None:
    """Tars and removes the already coalesced files

    :param filelist: List{str], list of files to be tarred and archived
    :param outfolder: str, folder name where the archive is to be stored
    :param state: SqCoalesceState, state of coalesceer
    :param dodel: bool, True if the coalesced files must be deleted
    :returns: Nothing
    """
    if filelist and outfolder:
        with tarfile.open(f'{outfolder}/_archive-{state.prefix}-'
                          f'{state.block_start}-{state.block_end}.tar.bz2',
                          'w:bz2') as f:
            for file in filelist:
                f.add(file)
    if dodel:
        [os.remove(x) for x in filelist]


def write_files(table: str, filelist: List[str], in_basedir: str,
                outfolder: str, partition_cols: List[str],
                state: SqCoalesceState, block_start, block_end) -> None:
    """Write the data from the list of files out as a single coalesced block

    We're fixing the compression in this function
    :param table: str, Name of the table for which we're writing the files
    :param filelist: List[str], list of files to write the data to
    :param in_basedir: str, base directory of the read files,
                       to get partition date
    :param outfolder: str, the outgoing folder to write the data to
    :param partition_cols: List[str], partition columns
    :param state: SqCoalesceState, coalescer state, for constructing filename
    :param block_start: dateime, starting time window of this coalescing block
    :param block_end: dateime, ending time window of this coalescing block
    :returns: Nothing
    """
    if not filelist and not state.schema.type == "record":
        return

    state.block_start = int(block_start.timestamp())
    state.block_end = int(block_end.timestamp())
    if filelist:
        this_df = ds.dataset(source=filelist, partitioning='hive',
                             partition_base_dir=in_basedir) \
            .to_table() \
            .to_pandas()
        state.wrrec_count += this_df.shape[0]

        if not this_df.empty:
            this_df = migrate_df(table, this_df, state.schema)

        if state.schema.type == "record":
            if not state.current_df.empty:
                this_df = this_df.set_index(state.keys)
                sett = set(this_df.index)
                setc = set(state.current_df.index)
                missing_set = setc.difference(sett)
                if missing_set:
                    missing_df = state.current_df.loc[missing_set]
                    this_df = pd.concat([this_df.reset_index(),
                                         missing_df.reset_index()])
                else:
                    this_df = this_df.reset_index()
    elif not state.current_df.empty:
        this_df = state.current_df.reset_index()
    else:
        return

    this_df.sqvers = state.schema.version  # Updating the schema version
    state.dbeng.write(state.table_name, "pandas", this_df, True,
                      state.schema.get_arrow_schema(),
                      state.pq_file_name)

    if state.schema.type == "record" and filelist:
        # Now replace the old dataframe with this new set for "record" types
        # Non-record types should never have current_df non-empty
        state.current_df = this_df.set_index(state.keys) \
                                  .drop(columns=['index'], errors='ignore') \
                                  .sort_values(by='timestamp') \
                                  .query('~index.duplicated(keep="last")')


def find_broken_files(parent_dir: str) -> List[str]:
    """Find any files in the parent_dir that pyarrow can't read
    :parame parent_dir: str, the parent directory to investigate
    :returns: list of broken files
    :rtype: list of strings
    """

    all_files = []
    broken_files = []
    ro, pa = os.path.split(parent_dir)
    for root, dirs, files in os.walk(parent_dir):
        if not '_archived' in root and not '.sq-coalescer.pid' in files and len(files) > 0:
            path = root.replace(ro, '')
            all_files.extend(list(map(lambda x: f"{path}/{x}", files)))
    for file in all_files:
        try:
            pq.ParquetFile(f"{ro}/{file}")
        except pq.lib.ArrowInvalid as e:
            broken_files.append(file)

    return broken_files


def move_broken_files(parent_dir: str, state: SqCoalesceState, out_dir: str = '_broken', ) -> None:
    """ based on the parent_directory, move any files that cannot be read by pyarrow
        to a safe directory to be investigated later
    :param parent_dir: str, the parent directory to investigate
    :param state: SqCoalesceState, needed for the logger
    :param out_dir: str, diretory to put the broken files in
    :returns: Nothing
    :rtype: None
    """

    broken_files = find_broken_files(parent_dir)
    ro, pa = os.path.split(parent_dir)

    for file in broken_files:
        src = f"{ro}/{file}"
        dst = f"{ro}/{out_dir}/{file}"

        if not os.path.exists(os.path.dirname(dst)):
            os.makedirs(os.path.dirname(dst))
        state.logger.debug(f"moving broken file {src} to {dst}")
        os.replace(src, dst)


def get_file_timestamps(filelist: List[str]) -> pd.DataFrame:
    """Read the files and construct a dataframe of files and timestamp of
       record in them.

    :param filelist: list, of full path name files, typically from pyarrow's
                     dataset.files
    :returns: dataframe of filename with the time it represents, sorted
    :rtype: pandas.DataFrame

    """
    if not filelist:
        return pd.DataFrame(columns=['file', 'timestamp'])

    # We can't rely on the system istat time to find the times involved
    # So read the data for each block and check. We tried using threading
    # and it didn't dramatically alter the results. Given that we might've
    # too many threads running with the poller and everything, we skipped
    # doing it.
    fname_list = []
    fts_list = []
    for file in filelist:
        try:
            ts = pd.read_parquet(file, columns=['timestamp'])
            fts_list.append(ts.timestamp.min())
            fname_list.append(file)
        except OSError:
            # skip this file because it can't be read, is probably 0 bytes
            logging.debug(f"not reading timestamp for {file}")

    # Construct file dataframe as its simpler to deal with
    if fname_list:
        fdf = pd.DataFrame({'file': fname_list, 'timestamp': fts_list})
        fdf['timestamp'] = humanize_timestamp(fdf.timestamp, 'UTC')
        return fdf.sort_values(by=['timestamp'])

    return pd.DataFrame(['file', 'timestamp'])


def migrate_df(table_name: str, df: pd.DataFrame,
               schema: SchemaForTable) -> pd.DataFrame:
    """Migrate the dataframe if necessary

    :param table_name: str, Name of the table
    :param df: pd.DataFrame, the dataframe to be migrated
    :param schema: SchemaForTable, the schema we're migrating to
    :returns: the migrated dataframe
    :rtype: pd.DataFrame
    """
    sqvers_list = df.sqvers.unique().tolist()
    for sqvers in sqvers_list:
        if sqvers != schema.version:
            migrate_fn = get_migrate_fn(table_name, sqvers,
                                        schema.version)
            if migrate_fn:
                df = migrate_fn(df)

    return df


def get_last_update_df(table_name: str, outfolder: str,
                       state: SqCoalesceState) -> pd.DataFrame:
    """Return a dataframe with the last known values for all keys
    The dataframe is sorted by timestamp and the index set to the keys. Works
    across namespaces.

    This is used when the coalesceer starts up and doesn't have any state
    about a table.

    :param table_name: str, name of the table we're getting data for
    :param outfolder: str, folder from where to gather the files
    :param state: SqCoalesceState, coalesceer state
    :returns: dataframe with the last known data for all hosts in namespace
    :rtype: pandas.DataFrame

    """
    dataset = ds.dataset(outfolder, partitioning='hive', format='parquet')
    files = sorted([x for x, y in zip(dataset.files, repeat(state.prefix))
                    if os.path.basename(x).startswith(y)])

    if not files:
        return pd.DataFrame()
    latest_filedict = {x.split('namespace=')[1].split('/')[0]: x
                       for x in files}

    latest_files = list(latest_filedict.values())
    current_df = ds.dataset(source=latest_files, partitioning='hive',
                            format='parquet') \
        .to_table() \
        .to_pandas(self_destruct=True)
    if not current_df.empty:
        # The data here could be old, so we need to migrate it first.
        current_df = migrate_df(table_name, current_df, state.schema)
        current_df = current_df.sort_values(by=['timestamp']) \
                               .set_index(state.keys)

    return current_df


def coalesce_resource_table(infolder: str, outfolder: str, archive_folder: str,
                            table: str, state: SqCoalesceState) -> None:
    """This routine coalesces all the parquet data in the folder provided

    This function MUST be called with sqPoller as the table the first time to
    build the polling period sample. Without this, its not possible to compute
    the records to be written for a period accurately. The polling periods are
    computed when this function is called the first time with None as the
    state field. This function stuffs the sqPoller timeblocks as the polling
    period in the state block and returns it. The state object returned also
    has some statistics written such as number of files written, number of
    records written and so on.

    :param infolder: str, folder to read data in from
    :param outfolder: str, folder to write data to
    :param archive_folder: str, folder to store the archived files in
    :param table: str, name of table we're coalesceing
    :param state: SqCoalesceState, state about this coalesceion run
    :returns: Nothing
    """

    def compute_block_start(start):
        return (start - timedelta(seconds=start.timestamp() %
                                  state.period.total_seconds()))

    partition_cols = ['sqvers', 'namespace']
    dodel = True

    if table == "sqPoller":
        wr_polling_period = True
        state.poller_periods = set()
    else:
        wr_polling_period = False
    state.wrfile_count = 0
    state.wrrec_count = 0
    state.table_name = table
    schema = state.schema

    if state.schema.type == "record":
        state.keys = schema.key_fields()
        if state.current_df.empty:
            state.current_df = get_last_update_df(table, outfolder, state)

    # Ignore reading the compressed files
    try:
        dataset = ds.dataset(infolder, partitioning='hive', format='parquet',
                             ignore_prefixes=state.ign_pfx)
    except OSError as e:
        move_broken_files(infolder, state=state)
        dataset = ds.dataset(infolder, partitioning='hive', format='parquet',
                             ignore_prefixes=state.ign_pfx)

    state.logger.info(f'Examining {len(dataset.files)} {table} files '
                      f'for coalescing')
    fdf = get_file_timestamps(dataset.files)
    if fdf.empty:
        if (table == 'sqPoller') or (not state.poller_periods):
            return

    # this is no longer true if we are skipping files that aren't readable
    #assert(len(dataset.files) == fdf.shape[0])
    polled_periods = sorted(state.poller_periods)
    if fdf.empty:
        state.logger.info(f'No updates for {table} to coalesce')
        start = polled_periods[0]
    else:
        start = fdf.timestamp.iloc[0]
    utcnow = datetime.now(timezone.utc)

    # We now need to determine if we're coalesceing a lot of data, at the start
    # or if we're only coalesceing for the last interval.
    if (utcnow < start):
        logging.error(
            'ERROR: Something is off, now is earlier than dates on files')
        return

    # We write data in fixed size time blocks. Data from 10 pm to 11 pm with a
    # period of 1 hour is written out as one block, data from 11-12 as another
    # and so on. Specifically, we write out 10:00:00 to 10:59:59 in the block
    block_start = compute_block_start(start)
    block_end = block_start + state.period

    # NOTE: You need the parentheses around the date comparison for some reason
    if (block_end > utcnow):
        return

    readblock = []
    wrfile_count = 0

    # We may start coalescing when nothing has changed for some initial period.
    # We have to write out records for that period.
    if schema.type == "record":
        for interval in polled_periods:
            if not fdf.empty and (block_end < interval):
                break
            pre_block_start = compute_block_start(interval)
            pre_block_end = pre_block_start + state.period
            write_files(table, readblock, infolder, outfolder, partition_cols,
                        state, pre_block_start, pre_block_end)

    for row in fdf.itertuples():
        if block_start <= row.timestamp < block_end:
            readblock.append(row.file)
            continue

        # Write data if either there's data to be written (readblock isn't
        # empty) OR this table is a record type and poller was alive during
        # this period (state's poller period for this window isn't blank
        if readblock or ((schema.type == "record") and
                         block_start in state.poller_periods):

            write_files(table, readblock, infolder, outfolder, partition_cols,
                        state, block_start, block_end)
            wrfile_count += len(readblock)
        if wr_polling_period and readblock:
            state.poller_periods.add(block_start)
        # Archive the saved files
        if readblock:
            archive_coalesced_files(readblock, archive_folder, state, dodel)

        # We have to find the timeslot where this record fits
        block_start = block_end
        block_end = block_start + state.period
        readblock = []
        if schema.type != "record":
            # We can jump directly to the timestamp corresonding to this
            # row's timestamp
            block_start = compute_block_start(row.timestamp)
            block_end = block_start + state.period
            if (row.timestamp > block_end) or (block_end > utcnow):
                break
            readblock = [row.file]
            continue

        while row.timestamp > block_end:
            if block_start in state.poller_periods:
                write_files(table, readblock, infolder, outfolder,
                            partition_cols, state, block_start, block_end)
                # Nothing to archive here, and we're not counting coalesced
                # records since these are duplicates
            block_start = block_end
            block_end = block_start + state.period
        if block_end > utcnow:
            break
        readblock = [row.file]

    # The last batch that ended before the block end
    if readblock or (fdf.empty and (schema.type == "record") and
                     block_start in state.poller_periods):
        write_files(table, readblock, infolder, outfolder, partition_cols,
                    state, block_start, block_end)
        wrfile_count += len(readblock)
        if wr_polling_period:
            state.poller_periods.add(block_start)
        archive_coalesced_files(readblock, archive_folder, state, dodel)

    state.wrfile_count = wrfile_count
    return



from suzieq.sqobjects.basicobj import SqObject
from suzieq.utils import convert_macaddr_format_to_colon


class ArpndObj(SqObject):

    def __init__(self, **kwargs):
        super().__init__(table='arpnd', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'ipAddress',
                                'columns', 'oif', 'macaddr', 'query_str']
        self._convert_args = {
            'macaddr': convert_macaddr_format_to_colon
        }



import pandas as pd
from ipaddress import ip_interface

from .engineobj import SqPandasEngine
from suzieq.utils import build_query_str, convert_macaddr_format_to_colon


class AddressObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'address'

    def addr_type(self, addr: list) -> list:

        rslt = []
        for a in addr:
            try:
                ipa = ip_interface(a)
                rslt.append(ipa._version)
            except ValueError:
                rslt.append(0)
        return rslt

    def get(self, **kwargs) -> pd.DataFrame:
        """Retrieve the dataframe that matches a given IPv4/v6/MAC address"""

        addr = kwargs.pop("address", [])
        columns = kwargs.get("columns", [])
        ipvers = kwargs.pop("ipvers", "")
        user_query = kwargs.pop('query_str', '')
        if user_query:
            if user_query.startswith('"') and user_query.endswith('"'):
                user_query = user_query[1:-1]

        vrf = kwargs.pop("vrf", "")
        addnl_fields = ['master']
        drop_cols = []

        v4addr = []
        v6addr = []
        macaddr = []
        try:
            addr_types = self.addr_type(addr)
        except ValueError:
            return pd.DataFrame({'error': ['Invalid address specified']})

        if columns != ['default'] and columns != ['*']:
            if ((4 in addr_types or ipvers == "v4") and
                    'ipAddressList' not in columns):
                addnl_fields.append('ipAddressList')
                drop_cols.append('ipAddressList')
            if ((6 in addr_types or ipvers == 'v6') and
                    'ip6AddressList' not in columns):
                addnl_fields.append('ip6AddressList')
                drop_cols.append('ip6AddressList')
            if ((0 in addr_types or ipvers == "l2") and
                    'macaddr' not in columns):
                addnl_fields.append('macaddr')
                drop_cols.append('macaddr')

        for i, a in enumerate(addr):
            if addr_types[i] == 0:
                # convert the macaddr format to internal format
                if '.' in a:
                    a = convert_macaddr_format_to_colon(a)
                macaddr.append(a.lower())
            elif addr_types[i] == 4:
                if '/' not in a:
                    a += '/'
                v4addr.append(a)
            elif addr_types[i] == 6:
                if '/' not in a:
                    a += '/'
                v6addr.append(a)

        if vrf == "default":
            master = ''
        else:
            master = vrf
        df = self.get_valid_df("address", master=master,
                               addnl_fields=addnl_fields, **kwargs)

        if df.empty:
            return df

        if vrf == "default":
            df = df.query('master==""')

        df = df.rename({'master': 'vrf'}, axis=1) \
            .replace({'vrf': {'': 'default'}})

        if 4 in addr_types:
            df = df.explode('ipAddressList').fillna({'ipAddressList': ''})

        if 6 in addr_types:
            df = df.explode('ip6AddressList').fillna({'ip6AddressList': ''})

        query_str = ''
        prefix = ''
        # IMPORTANT: Don't mess with this order of query. Some bug in pandas
        # prevents it from working if macaddr isn't first and your query
        # contains both a macaddr and an IP address.
        if macaddr:
            query_str += f'{prefix} macaddr.isin({macaddr}) '
            prefix = 'or'
        if v4addr:
            for a in v4addr:
                query_str += f'{prefix} ipAddressList.str.startswith("{a}") '
                prefix = 'or'
        if v6addr:
            for a in v6addr:
                query_str += f'{prefix} ip6AddressList.str.startswith("{a}") '
                prefix = 'or'

        if not query_str:
            if ipvers == "v4":
                query_str = 'ipAddressList.str.len() != 0'
            elif ipvers == "v6":
                query_str = 'ip6AddressList.str.len() != 0'
            elif ipvers == "l2":
                query_str == 'macaddr.str.len() != 0'

        if query_str:
            df = df.query(query_str)

        df = self._handle_user_query_str(df, user_query)
        return df.drop(columns=drop_cols)

    def unique(self, **kwargs) -> pd.DataFrame:
        """Specific here only to rename vrf column to master"""
        column = kwargs.pop("columns", None)
        if column == ["vrf"]:
            column = ["master"]
        df = super().unique(columns=column, **kwargs)
        if not df.empty:
            if 'vrf' in column:
                return df.query("master != 'bridge'") \
                         .rename({'master': 'vrf'}, axis='columns') \
                         .replace({'vrf': {'': 'default'}})
        return df

    def summarize(self, **kwargs) -> pd.DataFrame:
        """Summarize address related info"""

        self._init_summarize(self.iobj._table, **kwargs)
        if self.summary_df.empty:
            return self.summary_df

        self._add_field_to_summary('hostname', 'nunique', 'deviceCnt')
        self._add_field_to_summary('hostname', 'count', 'addressCnt')
        self._add_field_to_summary('macaddr', 'nunique', 'uniqueIfMacCnt')

        v6df = self.summary_df.explode('ip6AddressList') \
            .dropna(how='any') \
            .query('~ip6AddressList.str.startswith("fe80")')
        if not v6df.empty:
            v6device = v6df.groupby(by=['namespace'])['hostname'].nunique()
            v6addr = v6df.groupby(by=['namespace'])['ip6AddressList'].nunique()
            for i in self.ns.keys():
                self.ns[i].update(
                    {'deviceWithv6AddressCnt': v6device.get(i, 0)})
                self.ns[i].update({'uniqueV6AddressCnt': v6addr.get(i, 0)})
        else:
            for i in self.ns.keys():
                self.ns[i].update({'deviceWithv6AddressCnt': 0})
                self.ns[i].update({'uniqueV6AddressCnt': 0})

        v4df = self.summary_df.explode('ipAddressList') \
            .dropna(how='any') \
            .query('ipAddressList.str.len() != 0')
        if not v4df.empty:
            v4device = v4df.groupby(by=['namespace'])['hostname'].nunique()
            v4addr = v4df.groupby(by=['namespace'])['ipAddressList'].nunique()
            for i in self.ns.keys():
                self.ns[i].update(
                    {'deviceWithv4AddressCnt': v4device.get(i, 0)})
                self.ns[i].update({'uniqueV4AddressCnt': v4addr.get(i, 0)})

            v4df['prefixlen'] = v4df.ipAddressList.str.split('/').str[1]

            # this doesn't work if we've filtered by namespace
            #  pandas complains about an index problem
            #  so instead we have the more complicated expression below
            #  they are equivalent
            # v4pfx = v4df.groupby(by=['namespace'])['prefixlen'] \
            #             .value_counts().rename('count').reset_index()
            v4pfx = v4df.groupby(by=['namespace', 'prefixlen'],
                                 as_index=False)['ipAddressList'].count() \
                .dropna()
            v4pfx = v4pfx.rename(columns={'ipAddressList': 'count'})
            v4pfx['count'] = v4pfx['count'].astype(int)
            v4pfx = v4pfx.sort_values(by=['count'], ascending=False)

            for i in self.ns.keys():
                cnts = []
                v4pfx[v4pfx['namespace'] == i].apply(
                    lambda x: cnts.append(x['prefixlen']), axis=1, args=cnts)
                self.ns[i].update({'subnetsUsed': cnts})
                cnts = []
                v4pfx[v4pfx['namespace'] == i].apply(
                    lambda x: cnts.append({x['prefixlen']: x['count']}),
                    axis=1, args=cnts)
                self.ns[i].update({'subnetTopCounts': cnts[:3]})
        else:
            for i in self.ns.keys():
                self.ns[i].update({'deviceWithv4AddressCnt': 0})
                self.ns[i].update({'uniqueV4AddressCnt': 0})
                self.ns[i].update({'subnetsUsed': []})
                self.ns[i].update({'subnetTopCounts': []})

        self.summary_row_order = ['deviceCnt', 'addressCnt',
                                  'uniqueV4AddressCnt', 'uniqueV6AddressCnt',
                                  'uniqueIfMacCnt',
                                  'deviceWithv4AddressCnt',
                                  'deviceWithv6AddressCnt', 'subnetsUsed',
                                  'subnetTopCounts']
        self._post_summarize(check_empty_col='addressCnt')
        return self.ns_df.convert_dtypes()



import os
import re
from time import time
from typing import List
import logging
from pathlib import Path
import pyarrow.dataset as ds
from itertools import zip_longest
from datetime import datetime, timedelta, timezone
from contextlib import suppress
from shutil import rmtree
from collections import defaultdict

import pandas as pd
import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq

from suzieq.db.base_db import SqDB, SqCoalesceStats
from suzieq.utils import Schema, SchemaForTable

from .pq_coalesce import SqCoalesceState, coalesce_resource_table
from .migratedb import get_migrate_fn


class SqParquetDB(SqDB):

    def __init__(self, cfg: dict, logger: logging.Logger) -> None:
        '''Init the Parquet DB object'''
        self.cfg = cfg
        self.logger = logger or logging.getLogger()

    def supported_data_formats(self):
        return ['pandas']

    def get_tables(self):
        """Return list of tables known to parquet
        """
        folder = self._get_table_directory(None, False)
        dirs = Path(folder).glob('*')
        tables = [str(x.stem) for x in dirs
                  if not str(x.stem).startswith(('_', 'coalesced'))]

        return tables

    def read(self, table_name: str, data_format: str,
             **kwargs) -> pd.DataFrame:
        """Read the data specified from parquet files and return

        This function also implements predicate pushdown to filter the data
        as specified by the provided filters.

        :param table_name: str, the name of the table to be read
        :param data_format: str, Format the data's to be returned in,
                            (only pandas supported at this point)
        :param columns: List[str], list of columns requested to be read,
                        only those specified are returned, keyword arg
        :param key_fields: List[str], key fields for table, required to
                           deduplicate, keyword arg only
        :param view: str, one of ["latest", "all"], keyword arg only
        :param start: float, starting time window for data, timestamp,
                      can be 0 to indicate latest, keyword arg only
        :param end: float, ending time window for data, timestamp,
                    can be 0 to indicate latest, keyword arg only,
        :param kwargs: dict, the optional keyword arguments, addnl_filter,
                       and merge_fields, not needed typically
        :returns: pandas dataframe of the data specified, or None if
                  unsupported format
        :rtype: pd.DataFrame

        """

        if data_format not in self.supported_data_formats():
            return None

        start = kwargs.pop("start_time")
        end = kwargs.pop("end_time")
        view = kwargs.pop("view")
        fields = kwargs.pop("columns")
        key_fields = kwargs.pop("key_fields")
        addnl_filter = kwargs.pop("add_filter", None)
        merge_fields = kwargs.pop('merge_fields', {})
        hostname = kwargs.pop('hostname', [])
        namespace = kwargs.pop('namespace', [])

        folder = self._get_table_directory(table_name, False)

        if addnl_filter:
            # This is for special cases that are specific to an object
            query_str = addnl_filter
        else:
            query_str = None

        if query_str is None:
            # Make up a dummy query string to avoid if/then/else
            query_str = "timestamp != 0"

        # If sqvers is in the requested data, we've to handle it separately
        if 'sqvers' in fields:
            fields.remove('sqvers')
            need_sqvers = True
            max_vers = 0
        else:
            need_sqvers = False

        # If requesting a specific version of the data, handle that diff too
        sqvers = kwargs.pop('sqvers', None)
        datasets = []
        try:
            dirs = Path(folder)
            try:
                for elem in dirs.iterdir():
                    # Additional processing around sqvers filtering and data
                    if 'sqvers=' not in str(elem):
                        continue
                    if sqvers and f'sqvers={sqvers}' != elem:
                        continue
                    elif need_sqvers:
                        vers = float(str(elem).split('=')[-1])
                        if vers > max_vers:
                            max_vers = vers

                    datasets.append(ds.dataset(elem, format='parquet',
                                               partitioning='hive'))
            except FileNotFoundError:
                pass
            except Exception as e:
                raise e

            # Now find the exact set of files we need to go over
            cp_dataset = self._get_cp_dataset(table_name, need_sqvers, sqvers,
                                              view, start, end)
            if cp_dataset:
                datasets.append(cp_dataset)

            if not datasets:
                datasets = [ds.dataset(folder, format='parquet',
                                       partitioning='hive')]

            # Build the filters for predicate pushdown
            master_schema = self._build_master_schema(datasets)

            avail_fields = list(filter(lambda x: x in master_schema.names,
                                       fields))

            filters = self.build_ds_filters(
                start, end, master_schema, merge_fields=merge_fields, **kwargs)

            filtered_datasets = self._get_filtered_fileset(
                datasets, namespace, hostname)

            final_df = filtered_datasets \
                .to_table(filter=filters, columns=avail_fields) \
                .to_pandas(self_destruct=True) \
                .query(query_str) \
                .sort_values(by='timestamp')

            if merge_fields:
                # These are key fields that need to be set right before we do
                # the drop duplicates to avoid missing out all the data
                for field in merge_fields:
                    newfld = merge_fields[field]
                    if (field in final_df.columns and
                            newfld in final_df.columns):
                        final_df[newfld] = np.where(final_df[newfld],
                                                    final_df[newfld],
                                                    final_df[field])
                    elif (field in final_df.columns and
                          newfld not in final_df.columns):
                        final_df = final_df.rename(columns={field: newfld})

            # Because of how coalescing works, we can have multiple duplicated
            # entries with same timestamp. Remove them
            dupts_keys = key_fields + ['timestamp']
            final_df = final_df.set_index(dupts_keys) \
                .query('~index.duplicated(keep="last")') \
                .reset_index()
            if (not final_df.empty and (view == 'latest') and
                    all(x in final_df.columns for x in key_fields)):
                final_df = final_df.set_index(key_fields) \
                    .query('~index.duplicated(keep="last")')
        except (pa.lib.ArrowInvalid, OSError):
            return pd.DataFrame(columns=fields)

        if need_sqvers:
            final_df['sqvers'] = max_vers
            fields.insert(0, 'sqvers')

        cols = set(final_df.columns.tolist() + final_df.index.names)
        fields = [x for x in fields if x in cols]
        return final_df.reset_index()[fields]

    def write(self, table_name: str, data_format: str,
              data, coalesced: bool, schema: pa.lib.Schema,
              filename_cb, **kwargs) -> int:
        """Write the data supplied as a dataframe as a parquet file

        :param cfg: Suzieq configuration
        :param table_name: str, Name of the table to write data to
        :param data: data to be written, usually pandas DF, but can be
                     engine specific (spark, dask etc.)
        :param data_format: str, Format the data's to be returned in,
                            (only pandas supported at this point)
        :param coalesced: bool, True if data being written is in compacted form
        :param schema: pa.Schema, the schema for the data
        :param filename_cb: callable, callback function to create the filename
        :returns: status of write
        :rtype: integer

        """
        folder = self._get_table_directory(table_name, coalesced)
        if coalesced:
            partition_cols = ['sqvers', 'namespace']
        else:
            partition_cols = ['sqvers', 'namespace', 'hostname']

        schema_def = dict(zip(schema.names, schema.types))
        defvals = self._get_default_vals()

        if data_format == "pandas":
            if isinstance(data, pd.DataFrame):
                cols = data.columns

                # Ensure all fields are present
                for field in schema_def:
                    if field not in cols:
                        data[field] = defvals.get(schema_def[field], '')

                table = pa.Table.from_pandas(data, schema=schema,
                                             preserve_index=False)
            elif isinstance(data, pa.Table):
                table = data
            elif isinstance(data, dict):
                df = pd.DataFrame.from_dict(data["records"])
                table = pa.Table.from_pandas(df, schema=schema,
                                             preserve_index=False)

            if filename_cb:
                pq.write_to_dataset(table, root_path=folder,
                                    partition_cols=partition_cols,
                                    version="2.0", compression="ZSTD",
                                    partition_filename_cb=filename_cb,
                                    row_group_size=100000)
            else:
                pq.write_to_dataset(table, root_path=folder,
                                    partition_cols=partition_cols,
                                    version="2.0", compression="ZSTD",
                                    row_group_size=100000)

        return 0

    def coalesce(self, tables: List[str] = None, period: str = '',
                 ign_sqpoller: bool = False) -> None:
        """Coalesce all the resource parquet files in specified folder.

        This routine does not run periodically. It runs once and returns.

        :param tables: List[str], List of specific tables to coalesce,
                       empty for all
        :param period: str, coalescing period, needed for various internal
                       stuff
        :param ign_sqpoller: True if its OK to ignore the absence of sqpoller
                             to coalesce
        :returns: coalesce statistics list, one per table
        :rtype: SqCoalesceStats
        """

        infolder = self.cfg['data-directory']
        outfolder = self._get_table_directory('', True)  # root folder
        archive_folder = self.cfg.get('coalescer', {}) \
            .get('archive-directory',
                 f'{infolder}/_archived')

        if not period:
            period = self.cfg.get(
                'coalesceer', {'period': '1h'}).get('period', '1h')
        schemas = Schema(self.cfg.get('schema-directory'))
        state = SqCoalesceState(self.logger, period)

        state.logger = self.logger
        # Trying to be complete here. the ignore prefixes assumes you have
        # coalesceers across multiple time periods running, and so we need
        # to ignore the files created by the longer time period coalesceions.
        # In other words, weekly coalesceer should ignore monthly and yearly
        # coalesced files, monthly coalesceer should ignore yearly coalesceer
        # and so on.
        try:
            timeint = int(period[:-1])
            time_unit = period[-1]
            if time_unit == 'm':
                run_int = timedelta(minutes=timeint)
                state.prefix = 'sqc-m-'
                state.ign_pfx = ['.', '_', 'sqc-']
            if time_unit == 'h':
                run_int = timedelta(hours=timeint)
                state.prefix = 'sqc-h-'
                state.ign_pfx = ['.', '_', 'sqc-y-', 'sqc-d-', 'sqc-w-',
                                 'sqc-M-']
            elif time_unit == 'd':
                run_int = timedelta(days=timeint)
                if timeint > 364:
                    state.prefix = 'sqc-y-'
                    state.ign_pfx = ['.', '_', 'sqc-y-']
                elif timeint > 29:
                    state.prefix = 'sqc-M-'
                    state.ign_pfx = ['.', '_', 'sqc-M-', 'sqc-y-']
                else:
                    state.prefix = 'sqc-d-'
                    state.ign_pfx = ['.', '_', 'sqc-m-', 'sqc-d-', 'sqc-w-',
                                     'sqc-M-', 'sqc-y-']
            elif time_unit == 'w':
                run_int = timedelta(weeks=timeint)
                state.prefix = 'sqc-w-'
                state.ign_pfx = ['.', '_', 'sqc-w-', 'sqc-m-', 'sqc-y-']
            else:
                logging.error(f'Invalid unit for period, {time_unit}, '
                              'must be one of m/h/d/w')
        except ValueError:
            logging.error(f'Invalid time, {period}')
            return

        state.period = run_int
        # Create list of tables to coalesce.
        # TODO: Verify that we're only coalescing parquet tables here
        if tables:
            tables = [x for x in tables
                      if schemas.tables() and
                      (schemas.type_for_table(x) != "derivedRecord")]
        else:
            tables = [x for x in schemas.tables()
                      if schemas.type_for_table(x) != "derivedRecord"]
        if 'sqPoller' not in tables and not ign_sqpoller:
            # This is an error. sqPoller keeps track of discontinuities
            # among other things.
            self.logger.error(
                'No sqPoller data, cannot compute discontinuities')
            return
        else:
            # We want sqPoller to be first to compute discontinuities
            with suppress(ValueError):
                tables.remove('sqPoller')
            if not ign_sqpoller:
                tables.insert(0, 'sqPoller')

        # We've forced the sqPoller to be always the first table to coalesce
        stats = []
        for entry in tables:
            table_outfolder = f'{outfolder}/{entry}'
            table_infolder = f'{infolder}//{entry}'
            if archive_folder:
                table_archive_folder = f'{archive_folder}/{entry}'
            else:
                table_archive_folder = None
            state.current_df = pd.DataFrame()
            state.dbeng = self
            state.schema = SchemaForTable(entry, schemas, None)
            if not os.path.isdir(table_infolder):
                self.logger.info(
                    f'No input records to coalesce for {entry}')
                continue
            try:
                if not os.path.isdir(table_outfolder):
                    os.makedirs(table_outfolder)
                if (table_archive_folder and
                        not os.path.isdir(table_archive_folder)):
                    os.makedirs(table_archive_folder, exist_ok=True)
                # Migrate the data if needed
                self.logger.debug(f'Migrating data for {entry}')
                self.migrate(entry, state.schema)
                self.logger.debug(f'Migrating data for {entry}')
                start = time()
                coalesce_resource_table(table_infolder, table_outfolder,
                                        table_archive_folder, entry,
                                        state)
                end = time()
                self.logger.info(
                    f'coalesced {state.wrfile_count} files/{state.wrrec_count} '
                    f'records of {entry}')
                stats.append(SqCoalesceStats(entry, period, int(end-start),
                                             state.wrfile_count,
                                             state.wrrec_count,
                                             int(datetime.now(tz=timezone.utc)
                                                 .timestamp() * 1000)))
            except Exception:
                self.logger.exception(f'Unable to coalesce table {entry}')
                stats.append(SqCoalesceStats(entry, period, int(end-start),
                                             0, 0,
                                             int(datetime.now(tz=timezone.utc)
                                                 .timestamp() * 1000)))

        return stats

    def migrate(self, table_name: str, schema: SchemaForTable) -> None:
        """Migrates the data for the table specified to latest version

        :param table_name: str, The name of the table to migrate
        :param schema: SchemaForTable, the current schema
        :returns: None
        :rtype:
        """

        current_vers = schema.version
        defvals = self._get_default_vals()
        arrow_schema = schema.get_arrow_schema()
        schema_def = dict(zip(arrow_schema.names, arrow_schema.types))

        for sqvers in self._get_avail_sqvers(table_name, True):
            if sqvers != current_vers:
                migrate_rtn = get_migrate_fn(table_name, sqvers, current_vers)
                if migrate_rtn:
                    dataset = self._get_cp_dataset(table_name, True, sqvers,
                                                   'all', '', '')
                    for item in dataset.files:
                        try:
                            namespace = item.split('namespace=')[1] \
                                .split('/')[0]
                        except IndexError:
                            # Don't convert data not in our template
                            continue

                        df = pd.read_parquet(item)
                        df['sqvers'] = sqvers
                        df['namespace'] = namespace
                        newdf = migrate_rtn(df)

                        cols = newdf.columns
                        # Ensure all fields are present
                        for field in schema_def:
                            if field not in cols:
                                newdf[field] = defvals.get(schema_def[field],
                                                           '')

                        newdf.drop(columns=['namespace', 'sqvers'])

                        newitem = item.replace(f'sqvers={sqvers}',
                                               f'sqvers={current_vers}')
                        newdir = os.path.dirname(newitem)
                        if not os.path.exists(newdir):
                            os.makedirs(newdir, exist_ok=True)

                        table = pa.Table.from_pandas(
                            newdf, schema=schema.get_arrow_schema(),
                            preserve_index=False)
                        pq.write_to_dataset(table, newitem,
                                            version="2.0", compression="ZSTD",
                                            row_group_size=100000)
                        self.logger.debug(
                            f'Migrated {item} version {sqvers}->{current_vers}')
                        os.remove(item)

                    rmtree(f'{self._get_table_directory(table_name, True)}/sqvers={sqvers}',
                           ignore_errors=True)
        return

    def _get_avail_sqvers(self, table_name: str, coalesced: bool) -> List[str]:
        """Get list of DB versions for a given table.

        At this time it does not check if the DB is empty of not of these
        versions.

        :param table_name: str, name of table for which you want the versions
        :param coalesced: boolean, True if you want to look in coalesced dir
        :returns: list of DB versions
        :rtype: List[str]

        """

        folder = self._get_table_directory(table_name, coalesced)
        dirs = Path(folder)
        sqvers_list = []
        for folder in dirs.glob('sqvers=*'):
            with suppress(IndexError):
                sqvers = folder.name.split('sqvers=')[1]
                if sqvers:
                    sqvers_list.append(sqvers)

        return sqvers_list

    def _get_cp_dataset(self, table_name: str, need_sqvers: bool,
                        sqvers: str, view: str, start_time: float,
                        end_time: float) -> ds.dataset:
        """Get the list of files to read in coalesced dir

        This iterates over the coalesced files that need to be read and comes
        up with a list of files that corresponds to the timeslot the user has
        specified

        :param table_name: str, Table for which coalesced info is requested
        :param need_sqvers: bool, True if the user has requested that we
                            return the sqvers
        :param sqvers: str, if we're looking only for files of a specific vers
        :param view: str, whether to return the latest only OR all
        :param start_time: float, the starting time window of data needed
        : param end_time: float, the ending time window of data needed
        :returns: pyarrow dataset for the files to be read
        :rtype: pyarrow.dataset.dataset

        """

        filelist = []
        max_vers = 0

        folder = self._get_table_directory(table_name, True)

        if start_time and end_time or (view == "all"):
            # Enforcing the logic we have: if both start_time & end_time
            # are given, return all files since the model is that the user is
            # expecting to see all changes in the time window. Otherwise, the user
            # is expecting to see only the latest before an end_time OR after a
            # start_time.
            all_files = True
        else:
            all_files = False

        # We need to iterate otherwise the differing schema from different dirs
        # causes the read to abort.
        dirs = Path(folder)
        if not dirs.exists() or not dirs.is_dir():
            return

        for elem in dirs.iterdir():
            # Additional processing around sqvers filtering and data
            if 'sqvers=' not in str(elem):
                continue
            if sqvers and f'sqvers={sqvers}' != elem.name:
                continue
            elif need_sqvers:
                vers = float(str(elem).split('=')[-1])
                if vers > max_vers:
                    max_vers = vers

            dataset = ds.dataset(elem, format='parquet', partitioning='hive')
            total_files = len(dataset.files)

            # 100 is an arbitrary number, trying to reduce file I/O
            if all_files:
                files = dataset.files
            else:
                lists = defaultdict(list)
                for f in dataset.files:
                    nsp = os.path.dirname(f).split('namespace=')[-1]
                    lists[nsp].append(f)

                files = []
                for ele in lists:
                    lists[ele].sort()
                    if not start_time and not end_time:
                        files.append(lists[ele][-1])
                        continue

                    start_selected = False
                    for i, file in enumerate(lists[ele]):
                        thistime = os.path.basename(file).split('.')[0] \
                            .split('-')[-2:]
                        thistime = [int(x)*1000 for x in thistime]  # to msec
                        if (not start_time) or start_selected or (
                                thistime[0] <= start_time <= thistime[1]):
                            if not end_time:
                                files.extend(lists[ele][i:])
                                break
                            elif thistime[0] < end_time:
                                files.append(file)
                                start_selected = True
                            else:
                                break
            if files:
                filelist.extend(files)

        if filelist:
            return ds.dataset(filelist, format='parquet', partitioning='hive')
        else:
            return []

    def _build_master_schema(self, datasets: list) -> pa.lib.Schema:
        """Build the master schema from the list of diff versions
        We use this to build the filters and use the right type-based check
        for a field.
        """
        msch = datasets[0].schema
        msch_set = set(msch)
        for dataset in datasets[1:]:
            sch = dataset.schema
            sch_set = set(sch)
            if msch_set.issuperset(sch):
                continue
            elif sch_set.issuperset(msch):
                msch = sch
            else:
                for fld in sch_set-msch_set:
                    msch.append(fld)

        return msch

    def _get_filtered_fileset(self, datasets: list, namespace: list,
                              hostname: list) -> ds:
        """Filter the dataset based on the namespace and hostname

        We can use this method to filter out namespaces and hostnames based
        on regexes as well just regular strings.
        Args:
            datasets (list)): The datasets list incl coalesced and not files
            namespace (list): list of namespace strings
            hostname (list): list of hostname strings

        Returns:
            ds: pyarrow dataset of only the files that match filter
        """
        filelist = []
        for dataset in datasets:
            if not namespace:
                filelist.extend(dataset.files)
            else:
                ns_filelist = []
                for ns in namespace or []:
                    if ns.startswith('!'):
                        ns = ns[1:]
                        ns_filelist.extend([x for x in dataset.files
                                            if not re.search(f'namespace={ns}/',
                                                             x)])
                    else:
                        ns_filelist.extend([x for x in dataset.files
                                            if re.search(f'namespace={ns}/', x)])
                filelist.extend(ns_filelist)

            host_filelist = []
            for hn in hostname or []:
                if hn.startswith('!'):
                    hn = hn[1:]
                    host_filelist.extend([x for x in filelist
                                         if 'coalesced' in x or
                                         (not re.search(f'hostname={hn}/', x))])
                else:
                    host_filelist.extend([x for x in filelist
                                          if 'coalesced' in x or
                                          re.search(f'hostname={hn}/', x)])
            if hostname:
                filelist = host_filelist

        return ds.dataset(filelist, format='parquet', partitioning='hive')

    def _cons_int_filter(self, keyfld: str, filter_str: str) -> ds.Expression:
        '''Construct Integer filters with arithmetic operations'''
        if not isinstance(filter_str, str):
            return (ds.field(keyfld) == int(filter_str))

        # Check if we have logical operator (<, >, = etc.)
        if filter_str.startswith('<='):
            return (ds.field(keyfld) <= int(filter_str[2:]))
        elif filter_str.startswith('>='):
            return (ds.field(keyfld) >= int(filter_str[2:]))
        elif filter_str.startswith('<'):
            return (ds.field(keyfld) < int(filter_str[1:]))
        elif filter_str.startswith('>'):
            return (ds.field(keyfld) > int(filter_str[1:]))
        else:
            return (ds.field(keyfld) == int(filter_str))

    def build_ds_filters(self, start_tm: float, end_tm: float,
                         schema: pa.lib.Schema,
                         **kwargs) -> ds.Expression:
        """The new style of filters using dataset instead of ParquetDataset"""

        merge_fields = kwargs.pop('merge_fields', {})
        # The time filters first
        if start_tm and not end_tm:
            filters = ds.field("timestamp") >= start_tm
        elif end_tm and not start_tm:
            filters = ds.field("timestamp") <= end_tm
        elif start_tm and end_tm:
            filters = ((ds.field("timestamp") >= start_tm)
                       & (ds.field("timestamp") <= end_tm))
        else:
            filters = (ds.field("timestamp") != 0)

        sch_fields = schema.names
        for k, v in kwargs.items():
            if not v:
                continue
            if k not in sch_fields:
                self.logger.warning(f'Ignoring invalid field {k} in filter')
                continue

            ftype = schema.field(k).type
            if k in merge_fields:
                k = merge_fields[k]

            if isinstance(v, list):
                infld = []
                notinfld = []
                or_filters = None
                for e in v:
                    if isinstance(e, str) and e.startswith("!"):
                        if ftype == 'int64':
                            notinfld.append(int(e[1:]))
                        else:
                            notinfld.append(e[1:])
                    else:
                        if ftype == 'int64':
                            if or_filters is not None:
                                or_filters = or_filters | \
                                    self._cons_int_filter(k, e)
                            else:
                                or_filters = self._cons_int_filter(k, e)
                        else:
                            infld.append(e)
                if infld and notinfld:
                    filters = filters & (ds.field(k).isin(infld) &
                                         ~ds.field(k).isin(notinfld))
                elif infld:
                    filters = filters & (ds.field(k).isin(infld))
                elif notinfld:
                    filters = filters & (~ds.field(k).isin(notinfld))

                if or_filters is not None:
                    filters = filters & (or_filters)
            else:
                if isinstance(v, str) and v.startswith("!"):
                    if ftype == 'int64':
                        filters = filters & (ds.field(k) != int(v[1:]))
                    else:
                        filters = filters & (ds.field(k) != v[1:])
                else:
                    if ftype == 'int64':
                        filters = filters & self._cons_int_filter(k, v)
                    else:
                        filters = filters & (ds.field(k) == v)

        return filters

    def _get_table_directory(self, table_name: str, coalesced: bool) -> str:
        """Return the directory name for the table specified

        :param table_name: str, The table for which you want the folder
        :param coalesced: bool, True if you want the folder for the coalesced
                          data
        :returns: folder for the specified table name
        :rtype: str

        """
        if coalesced:
            dir = self.cfg.get('coalescer', {})\
                .get('coalesce-directory',
                     f'{self.cfg.get("data-directory")}/coalesced')
        else:
            dir = f'{self.cfg.get("data-directory")}'

        if table_name:
            return f'{dir}/{table_name}'
        else:
            return dir

    def _get_default_vals(self) -> dict:
        return({
            pa.string(): "",
            pa.int32(): 0,
            pa.int64(): 0,
            pa.float32(): 0.0,
            pa.float64(): 0.0,
            pa.date64(): 0.0,
            pa.bool_(): False,
            pa.list_(pa.string()): [],
            pa.list_(pa.int64()): [],
        })



import time
from nubia import command, argument
import pandas as pd

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.devconfig import DevconfigObj


@command("devconfig", help="Act on device data")
class DevconfigCmd(SqCommand):
    """Device configurations"""

    def __init__(
            self,
            engine: str = "pandas",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",
            query_str: str = " ",
            columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=DevconfigObj,
        )

    @command("show", help="Show device information")
    def show(self):
        """Show device config info
        """
        if self.columns is None:
            return

        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        df = self._invoke_sqobj(self.sqobj.get,
                                hostname=self.hostname, columns=self.columns,
                                namespace=self.namespace,
                                query_str=self.query_str,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        if not self.format or (self.format == 'text'):
            self.format = 'devconfig'
        return self._gen_output(df)

    @command("unique", help="Show unique information about columns")
    def unique(self):
        """
        Unique device config info
        """

        df = pd.DataFrame(
            {'error': ['Unique not supported for Device Config']})
        return self._gen_output(df)



import time
import pandas as pd
from nubia import command, argument

from suzieq.utils import SchemaForTable
from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.topology import TopologyObj


@command("topology", help="build and act on topology data")
class TopologyCmd(SqCommand):
    """Information about the topology constructed from various protocols"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",
        query_str: str = ' ',
        columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=TopologyObj
        )

    @command("show")
    @argument("polled", description="Is the device polled by Suzieq",
              choices=['True', 'False'])
    @argument("ifname", description="interface name to qualify")
    @argument("via", description="filter the method by which topology is seen",
              choices=['arpnd', 'bgp', 'lldp', 'ospf'])
    @argument("peerHostname",
              description="filter the result by specified peerHostname")
    def show(self, polled: str = '', ifname: str = '', via: str = '',
             peerHostname: str = ''):
        """Show table of topology information"""
        # Get the default display field names
        if self.columns is None:
            return

        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        try:
            df = self.sqobj.get(
                namespace=self.namespace,
                ifname=ifname.split(),
                via=via.split(),
                polled=polled,
                peerHostname=peerHostname.split(),
                hostname=self.hostname,
                query_str=self.query_str,
            )
        except Exception as e:
            df = pd.DataFrame({'error': ['ERROR: {}'.format(str(e))]})

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        cols = SchemaForTable('topology', self.schemas).sorted_display_fields()
        cols = [x for x in cols if x in df.columns]
        return self._gen_output(df[cols])

    @command("summarize")
    @argument("via", description="filter the method by which topology is seen",
              choices=['arpnd', 'bgp', 'lldp', 'ospf'])
    def summarize(self, via: str = ""):
        """Summarize topology information"""
        # Get the default display field names
        if self.columns is None:
            return

        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        try:
            df = self.sqobj.summarize(
                hostname=self.hostname,
                namespace=self.namespace,
                via=via.split(),
            )
        except Exception as e:
            df = pd.DataFrame({'error': ['ERROR: {}'.format(str(e))]})

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        if not df.empty:
            return self._gen_output(df)



import pandas as pd
import numpy as np

from suzieq.sqobjects.basicobj import SqObject
from suzieq.utils import SchemaForTable, humanize_timestamp


class OspfObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='ospf', **kwargs)
        self._addnl_fields = ['passive', 'area', 'state']
        self._addnl_nbr_fields = ['state']
        self._valid_get_args = ['namespace', 'hostname', 'columns',
                                'vrf', 'ifname', 'state', 'query_str']
        self._valid_assert_args = ['namespace', 'vrf', 'status']
        self._valid_arg_vals = {
            'state': ['full', 'other', 'passive', ''],
            'status': ['all', 'pass', 'fail'],
        }

    def aver(self, **kwargs):
        """Assert that the OSPF state is OK"""

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')

        try:
            self.validate_assert_input(**kwargs)
        except Exception as error:
            df = pd.DataFrame({'error': [f'{error}']})
            return df

        return self.engine.aver(**kwargs)

    def humanize_fields(self, df: pd.DataFrame, subset=None) -> pd.DataFrame:
        '''Humanize the timestamp and boot time fields'''
        if df.empty:
            return df

        if 'lastChangeTime' in df.columns:
            df['lastChangeTime'] = humanize_timestamp(
                df.lastChangeTime.fillna(0),
                self.cfg.get('analyzer', {}).get('timezone', None))

            if 'adjState' in df.columns:
                df['lastChangeTime'] = np.where(df.adjState == "passive",
                                                pd.Timestamp(0),
                                                df.lastChangeTime)

        return df
